from collections import Counter
from pathlib import Path

import numpy as np
import pytest


@pytest.fixture
def file_systems(tmp_path: Path):
    import pyarrow.fs as fs

    from deltalake.fs import DeltaStorageHandler

    store = fs.PyFileSystem(DeltaStorageHandler(str(tmp_path.absolute())))
    arrow_fs = fs.SubTreeFileSystem(str(tmp_path.absolute()), fs.LocalFileSystem())
    return (store, arrow_fs)


@pytest.fixture
def table_data():
    import pyarrow as pa

    return pa.Table.from_arrays(
        [pa.array([1, 2, 3]), pa.array(["a", "b", "c"])], names=["int", "str"]
    )


@pytest.mark.pyarrow
def test_file_info(file_systems, table_data):
    store, arrow_fs = file_systems
    file_path = "table.parquet"

    import pyarrow.parquet as pq

    pq.write_table(table_data, file_path, filesystem=arrow_fs)

    info = store.get_file_info(file_path)
    arrow_info = arrow_fs.get_file_info(file_path)

    assert type(info) is type(arrow_info)
    assert info.path == arrow_info.path
    assert info.type == arrow_info.type
    assert info.size == arrow_info.size
    assert info.mtime_ns == arrow_info.mtime_ns
    assert info.mtime == arrow_info.mtime


@pytest.mark.pyarrow
def test_get_file_info_selector(file_systems):
    store, arrow_fs = file_systems
    import pyarrow as pa
    import pyarrow.dataset as ds
    import pyarrow.fs as fs

    table = pa.table({"a": range(10), "b": np.random.randn(10), "c": [1, 2] * 5})
    partitioning = ds.partitioning(pa.schema([("c", pa.int64())]), flavor="hive")
    ds.write_dataset(
        table,
        "/",
        partitioning=partitioning,
        format="parquet",
        filesystem=arrow_fs,
    )

    selector = fs.FileSelector("/", recursive=True)
    infos = store.get_file_info(selector)
    arrow_infos = arrow_fs.get_file_info(selector)

    assert Counter([i.type for i in infos]) == Counter([i.type for i in arrow_infos])


@pytest.mark.pyarrow
def test_open_input_file(file_systems, table_data):
    store, arrow_fs = file_systems
    file_path = "table.parquet"
    import pyarrow.parquet as pq

    pq.write_table(table_data, file_path, filesystem=arrow_fs)

    file = store.open_input_file(file_path)
    arrow_file = arrow_fs.open_input_file(file_path)

    # Check the metadata
    assert file.mode == arrow_file.mode
    assert file.closed == arrow_file.closed
    assert file.size() == arrow_file.size()
    assert file.isatty() == arrow_file.isatty()
    assert file.readable() == arrow_file.readable()
    assert file.seekable() == arrow_file.seekable()

    # Check reading the same content
    assert file.read() == arrow_file.read()
    # Check subsequent read (should return no data anymore)
    assert file.read() == arrow_file.read()
    assert file.read() == b""

    file = store.open_input_file(file_path)
    arrow_file = arrow_fs.open_input_file(file_path)

    # Check seeking works as expected
    assert file.tell() == arrow_file.tell()
    assert file.seek(2) == arrow_file.seek(2)
    assert file.tell() == arrow_file.tell()
    assert file.tell() == 2

    # check reading works as expected
    assert file.read(10) == arrow_file.read(10)
    assert file.read1(10) == arrow_file.read1(10)
    assert file.read_at(10, 0) == arrow_file.read_at(10, 0)


@pytest.mark.pyarrow
def test_open_input_file_with_size(tmp_path, table_data):
    file_path = "table.parquet"
    input_size = 12345  # incorrect file size for testing purposes
    import pyarrow.fs as fs
    import pyarrow.parquet as pq

    from deltalake.fs import DeltaStorageHandler

    # test that injected file size gets stored correctly
    store1 = DeltaStorageHandler(
        str(tmp_path.absolute()), known_sizes={file_path: input_size}
    )
    wrapped_fs = fs.PyFileSystem(store1)
    arrow_fs = fs.SubTreeFileSystem(str(tmp_path.absolute()), fs.LocalFileSystem())
    pq.write_table(table_data, file_path, filesystem=arrow_fs)
    file = wrapped_fs.open_input_file(file_path)
    assert file.size() == input_size

    # confirm that true size is different
    store2 = DeltaStorageHandler(str(tmp_path.absolute()))
    wrapped_fs = fs.PyFileSystem(store2)
    arrow_fs = fs.SubTreeFileSystem(str(tmp_path.absolute()), fs.LocalFileSystem())
    pq.write_table(table_data, file_path, filesystem=arrow_fs)
    file = wrapped_fs.open_input_file(file_path)
    assert file.size() != input_size


@pytest.mark.pyarrow
def test_read_table(file_systems, table_data):
    store, arrow_fs = file_systems
    file_path = "table.parquet"
    import pyarrow as pa
    import pyarrow.parquet as pq

    pq.write_table(table_data, file_path, filesystem=arrow_fs)

    table = pq.read_table(file_path, filesystem=store)
    arrow_table = pq.read_table(file_path, filesystem=arrow_fs)

    assert isinstance(table, pa.Table)
    assert table.equals(arrow_table)


@pytest.mark.pyarrow
def test_read_dataset(file_systems):
    store, arrow_fs = file_systems
    import pyarrow as pa
    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    table = pa.table({"a": range(10), "b": np.random.randn(10), "c": [1, 2] * 5})

    pq.write_table(table.slice(0, 5), "data1.parquet", filesystem=arrow_fs)
    pq.write_table(table.slice(5, 10), "data2.parquet", filesystem=arrow_fs)

    dataset = ds.dataset("/", format="parquet", filesystem=store)
    ds_table = dataset.to_table()

    assert table.schema == dataset.schema
    assert table.equals(ds_table)


@pytest.mark.pyarrow
def test_write_table(file_systems):
    store, _ = file_systems
    import pyarrow as pa
    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    table = pa.table({"a": range(10), "b": np.random.randn(10), "c": [1, 2] * 5})

    pq.write_table(table.slice(0, 5), "data1.parquet", filesystem=store)
    pq.write_table(table.slice(5, 10), "data2.parquet", filesystem=store)

    dataset = ds.dataset("/", format="parquet", filesystem=store)
    ds_table = dataset.to_table()

    assert table.schema == ds_table.schema
    assert table.equals(ds_table)


@pytest.mark.pyarrow
def test_write_partitioned_dataset(file_systems):
    store, arrow_fs = file_systems
    import pyarrow as pa
    import pyarrow.dataset as ds

    table = pa.table({"a": range(10), "b": np.random.randn(10), "c": [1, 2] * 5})

    partitioning = ds.partitioning(pa.schema([("c", pa.int64())]), flavor="hive")
    ds.write_dataset(
        table,
        "/",
        partitioning=partitioning,
        format="parquet",
        filesystem=store,
    )

    dataset = ds.dataset(
        "/", format="parquet", filesystem=arrow_fs, partitioning=partitioning
    )
    ds_table = dataset.to_table().select(["a", "b", "c"])

    dataset = ds.dataset(
        "/", format="parquet", filesystem=store, partitioning=partitioning
    )
    ds_table2 = dataset.to_table().select(["a", "b", "c"])

    assert table.schema == ds_table.schema
    assert table.schema == ds_table2.schema
    assert table.shape == ds_table.shape
    assert table.shape == ds_table2.shape



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pyarrow as pa


def get_spark():
    import delta
    import delta.pip_utils
    import delta.tables
    import pyspark

    builder = (
        pyspark.sql.SparkSession.builder.appName("MyApp")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config(
            "spark.sql.catalog.spark_catalog",
            "org.apache.spark.sql.delta.catalog.DeltaCatalog",
        )
        .config("spark.sql.ansi.enabled", "false")
    )
    return delta.pip_utils.configure_spark_with_delta_pip(builder).getOrCreate()


def assert_spark_read_equal(
    expected: "pa.Table", uri: str, sort_by: list[str] = ["int32"]
):
    from pandas.testing import assert_frame_equal

    spark = get_spark()
    df = spark.read.format("delta").load(uri)

    # Spark and pyarrow don't convert these types to the same Pandas values
    incompatible_types = ["timestamp", "struct"]

    assert_frame_equal(
        df.toPandas()
        .sort_values(sort_by, ignore_index=True)
        .drop(incompatible_types, axis="columns", errors="ignore"),
        expected.to_pandas()
        .sort_values(sort_by, ignore_index=True)
        .drop(incompatible_types, axis="columns", errors="ignore"),
    )


def run_stream_with_checkpoint(source_table: str):
    spark = get_spark()

    stream_path = source_table + "/stream"
    checkpoint_path = stream_path + "streaming_checkpoints/"

    streaming_df = spark.readStream.format("delta").load(source_table)
    query = (
        streaming_df.writeStream.format("delta")
        .outputMode("append")
        .option("checkpointLocation", checkpoint_path)
        .option("mergeSchema", "true")
        .start(stream_path)
    )
    query.processAllAvailable()
    query.stop()



from __future__ import annotations

import os
import pathlib
import subprocess
import time
from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from time import sleep
from typing import TYPE_CHECKING, Generator

import pytest
from arro3.core import Array, DataType, Field, Schema, Table
from azure.storage import blob

from deltalake import DeltaTable, WriterProperties, write_deltalake

if TYPE_CHECKING:
    import pyarrow as pa
    from minio import Minio


def wait_till_host_is_available(host: str, timeout_sec: int = 0.5):
    spacing = 2
    start = time.monotonic()
    while True:
        if time.monotonic() - start > timeout_sec:
            raise TimeoutError(f"Host {host} is not available after {timeout_sec} sec")

        try:
            subprocess.run(["curl", host], timeout=timeout_sec * 1000, check=True)
        except Exception:
            pass
        else:
            break

        sleep(spacing)


@pytest.fixture(scope="session")
def s3_localstack_creds():
    endpoint_url = "http://localhost:4566"

    config = dict(
        AWS_REGION="us-east-1",
        AWS_ACCESS_KEY_ID="deltalake",
        AWS_SECRET_ACCESS_KEY="weloverust",
        AWS_ENDPOINT_URL=endpoint_url,
    )

    env = os.environ.copy()
    env.update(config)

    setup_commands = [
        [
            "aws",
            "s3api",
            "create-bucket",
            "--bucket",
            "deltars",
            "--endpoint-url",
            endpoint_url,
        ],
        [
            "aws",
            "s3",
            "sync",
            "--quiet",
            "../crates/test/tests/data/simple_table",
            "s3://deltars/simple",
            "--endpoint-url",
            endpoint_url,
        ],
    ]

    wait_till_host_is_available(endpoint_url)

    try:
        for args in setup_commands:
            subprocess.run(args, env=env)
    except OSError:
        pytest.skip("aws cli not installed")

    yield config

    shutdown_commands = [
        [
            "aws",
            "s3",
            "rm",
            "--quiet",
            "--recursive",
            "s3://deltars",
            "--endpoint-url",
            endpoint_url,
        ],
        [
            "aws",
            "s3api",
            "delete-bucket",
            "--bucket",
            "deltars",
            "--endpoint-url",
            endpoint_url,
        ],
    ]

    for args in shutdown_commands:
        subprocess.run(args, env=env)


@pytest.fixture()
def s3_localstack(monkeypatch, s3_localstack_creds):
    monkeypatch.setenv("AWS_ALLOW_HTTP", "TRUE")
    for key, value in s3_localstack_creds.items():
        monkeypatch.setenv(key, value)


@pytest.fixture(scope="session")
def azurite_creds():
    # These are the well-known values
    # https://learn.microsoft.com/en-us/azure/storage/common/storage-use-azurite?tabs=visual-studio#well-known-storage-account-and-key
    account_name = "devstoreaccount1"
    config = dict(
        AZURE_STORAGE_ACCOUNT_NAME=account_name,
        AZURE_STORAGE_ACCOUNT_KEY="Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==",
        AZURE_STORAGE_CONTAINER_NAME="deltars",
        AZURE_STORAGE_USE_EMULATOR="true",
        AZURE_STORAGE_USE_HTTP="true",
        AZURE_STORAGE_ENDPOINT=f"http://localhost:10000/{account_name}",
    )

    env = os.environ.copy()
    env.update(config)
    conn_str = (
        "DefaultEndpointsProtocol=http;"
        f"AccountName={config['AZURE_STORAGE_ACCOUNT_NAME']};"
        f"AccountKey={config['AZURE_STORAGE_ACCOUNT_KEY']};"
        f"BlobEndpoint={config['AZURE_STORAGE_ENDPOINT']};"
    )
    env["AZURE_STORAGE_CONNECTION_STRING"] = conn_str
    wait_till_host_is_available(config["AZURE_STORAGE_ENDPOINT"])
    try:
        blob_client = blob.BlobServiceClient.from_connection_string(conn_str=conn_str)
        container = blob_client.create_container(
            name=config["AZURE_STORAGE_CONTAINER_NAME"]
        )
        yield config
    finally:
        container.delete_container()


@pytest.fixture()
def azurite_env_vars(monkeypatch, azurite_creds):
    for key, value in azurite_creds.items():
        monkeypatch.setenv(key, value)


@pytest.fixture(scope="session")
def azurite_sas_creds(azurite_creds):
    env = os.environ.copy()
    env.update(azurite_creds)
    env["AZURE_STORAGE_CONNECTION_STRING"] = (
        "DefaultEndpointsProtocol=http;"
        f"AccountName={azurite_creds['AZURE_STORAGE_ACCOUNT_NAME']};"
        f"AccountKey={azurite_creds['AZURE_STORAGE_ACCOUNT_KEY']};"
        f"BlobEndpoint={azurite_creds['AZURE_STORAGE_ENDPOINT']};"
    )
    sas_token = blob.generate_container_sas(
        account_name=azurite_creds["AZURE_STORAGE_ACCOUNT_NAME"],
        container_name=azurite_creds["AZURE_STORAGE_CONTAINER_NAME"],
        account_key=azurite_creds["AZURE_STORAGE_ACCOUNT_KEY"],
        permission=blob.ContainerSasPermissions(
            read=True,
            write=True,
            list=True,
            delete=True,
        ),
        expiry=datetime.now(tz=timezone.utc) + timedelta(hours=1),
        start=datetime.now(tz=timezone.utc),
        protocol="http",
    )
    creds = {key: value for key, value in azurite_creds.items() if "KEY" not in key}
    creds["SAS_TOKEN"] = sas_token

    return creds


@pytest.fixture()
def sample_data_pyarrow() -> "pa.Table":
    nrows = 5
    import pyarrow as pa

    return pa.table(
        {
            "utf8": pa.array([str(x) for x in range(nrows)]),
            "int64": pa.array(list(range(nrows)), pa.int64()),
            "int32": pa.array(list(range(nrows)), pa.int32()),
            "int16": pa.array(list(range(nrows)), pa.int16()),
            "int8": pa.array(list(range(nrows)), pa.int8()),
            "float32": pa.array([float(x) for x in range(nrows)], pa.float32()),
            "float64": pa.array([float(x) for x in range(nrows)], pa.float64()),
            "bool": pa.array([x % 2 == 0 for x in range(nrows)]),
            "binary": pa.array([str(x).encode() for x in range(nrows)]),
            "decimal": pa.array([Decimal("10.000") + x for x in range(nrows)]),
            "date32": pa.array(
                [date(2022, 1, 1) + timedelta(days=x) for x in range(nrows)]
            ),
            "timestamp": pa.array(
                [datetime(2022, 1, 1) + timedelta(hours=x) for x in range(nrows)]
            ),
            "struct": pa.array([{"x": x, "y": str(x)} for x in range(nrows)]),
            "list": pa.array(
                [list(range(x + 1)) for x in range(nrows)],
                type=pa.list_(pa.field("inner", pa.int64(), nullable=False)),
            ),
            # NOTE: https://github.com/apache/arrow-rs/issues/477
            #'map': pa.array([[(str(y), y) for y in range(x)] for x in range(nrows)], pa.map_(pa.string(), pa.int64())),
        }
    )


@pytest.fixture()
def existing_table(tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"):
    path = str(tmp_path)
    write_deltalake(path, sample_data_pyarrow)
    return DeltaTable(path)


@pytest.fixture()
def sample_table() -> Table:
    nrows = 5
    return Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                Field("id", type=DataType.string(), nullable=True),
            ),
            "price": Array(
                list(range(nrows)), Field("price", type=DataType.int64(), nullable=True)
            ),
            "sold": Array(
                list(range(nrows)), Field("sold", type=DataType.int32(), nullable=True)
            ),
            "deleted": Array(
                [False] * nrows, Field("deleted", type=DataType.bool(), nullable=True)
            ),
        },
    )


@pytest.fixture()
def existing_sample_table(tmp_path: pathlib.Path, sample_table: Table):
    path = str(tmp_path)
    write_deltalake(path, sample_table)
    return DeltaTable(path)


@pytest.fixture()
def sample_table_with_spaces_numbers() -> Table:
    nrows = 5
    return Table.from_pydict(
        {
            "1id": Array(["1", "2", "3", "4", "5"], DataType.string()),
            "price": Array(list(range(nrows)), DataType.int64()),
            "sold items": Array(list(range(nrows)), DataType.int32()),
            "deleted": Array(
                [False] * nrows,
                DataType.bool(),
            ),
        },
        schema=Schema(
            fields=[
                Field("1id", type=DataType.string(), nullable=True),
                Field("price", type=DataType.int64(), nullable=True),
                Field("sold items", type=DataType.int32(), nullable=True),
                Field("deleted", type=DataType.bool(), nullable=True),
            ]
        ),
    )


@pytest.fixture(scope="session")
def minio_container() -> Generator[tuple[dict, Minio], None, None]:
    """Start a MinIO test container for S3-compatible object storage."""
    from testcontainers.minio import MinioContainer

    container = MinioContainer(
        image="minio/minio:latest",
        access_key="minioadmin",
        secret_key="minioadmin",
    )
    container.start()

    client = container.get_client()

    container_config = container.get_config()

    config = {
        "AWS_REGION": "us-east-1",
        "AWS_ACCESS_KEY_ID": container_config["access_key"],
        "AWS_SECRET_ACCESS_KEY": container_config["secret_key"],
        "AWS_ENDPOINT_URL": "http://" + container_config["endpoint"],
        "AWS_ALLOW_HTTP": "TRUE",
        "AWS_S3_ALLOW_UNSAFE_RENAME": "true",
    }

    yield (config, client)

    container.stop()


@pytest.fixture()
def minio_s3_env(monkeypatch, minio_container):
    """Set up environment variables for MinIO S3-compatible storage."""
    for key, value in minio_container.items():
        monkeypatch.setenv(key, value)
    return minio_container


@pytest.fixture()
def writer_properties():
    return WriterProperties(compression="GZIP", compression_level=0)



import pathlib

import pytest
from arro3.core import DataType, Field, Schema

from deltalake import DeltaTable, write_deltalake
from deltalake.writer._conversion import _convert_arro3_schema_to_delta


@pytest.mark.parametrize(
    "input_schema,expected_schema",
    [
        # Basic types - identity
        (
            Schema(fields=[Field("foo", DataType.int64())]),
            Schema(fields=[Field("foo", DataType.int64())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.int32())]),
            Schema(fields=[Field("foo", DataType.int32())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.int16())]),
            Schema(fields=[Field("foo", DataType.int16())]),
        ),
        # Unsigned integers to signed
        (
            Schema(fields=[Field("foo", DataType.uint8())]),
            Schema(fields=[Field("foo", DataType.int8())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint16())]),
            Schema(fields=[Field("foo", DataType.int16())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint32())]),
            Schema(fields=[Field("foo", DataType.int32())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint64())]),
            Schema(fields=[Field("foo", DataType.int64())]),
        ),
        # Timestamps
        (
            Schema(fields=[Field("foo", DataType.timestamp("s"))]),
            Schema(fields=[Field("foo", DataType.timestamp("us"))]),
        ),
        (
            Schema(
                fields=[Field("foo", DataType.timestamp("ns", tz="Europe/Amsterdam"))]
            ),
            Schema(fields=[Field("foo", DataType.timestamp("us", tz="UTC"))]),
        ),
        # Nullability variations
        (
            Schema(fields=[Field("foo", DataType.uint16(), nullable=False)]),
            Schema(fields=[Field("foo", DataType.int16(), nullable=False)]),
        ),
        (
            Schema(fields=[Field("foo", DataType.timestamp("ns"), nullable=True)]),
            Schema(fields=[Field("foo", DataType.timestamp("us"), nullable=True)]),
        ),
        # List of unsigned ints
        (
            Schema(fields=[Field("foo", DataType.list(DataType.uint32()))]),
            Schema(fields=[Field("foo", DataType.list(DataType.int32()))]),
        ),
        (
            Schema(fields=[Field("foo", DataType.large_list(DataType.uint8()))]),
            Schema(fields=[Field("foo", DataType.large_list(DataType.int8()))]),
        ),
        # List with nullable inner fields
        (
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.uint8(), nullable=True)),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.int8(), nullable=True)),
                    )
                ]
            ),
        ),
        # List with non-nullable inner fields
        (
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.uint8(), nullable=False)),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.int8(), nullable=False)),
                    )
                ]
            ),
        ),
        # Deeply nested list
        (
            Schema(
                fields=[
                    Field(
                        "deep_list",
                        DataType.list(
                            Field(
                                "level_1",
                                DataType.list(
                                    Field(
                                        "level_2",
                                        DataType.list(
                                            Field(
                                                "value",
                                                DataType.uint16(),
                                                nullable=True,
                                            )
                                        ),
                                        nullable=True,
                                    )
                                ),
                                nullable=True,
                            )
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "deep_list",
                        DataType.list(
                            Field(
                                "level_1",
                                DataType.list(
                                    Field(
                                        "level_2",
                                        DataType.list(
                                            Field(
                                                "value", DataType.int16(), nullable=True
                                            )
                                        ),
                                        nullable=True,
                                    )
                                ),
                                nullable=True,
                            )
                        ),
                    )
                ]
            ),
        ),
        # Fixed-size list
        (
            Schema(fields=[Field("foo", DataType.list(DataType.uint16(), 5))]),
            Schema(fields=[Field("foo", DataType.list(DataType.int16(), 5))]),
        ),
        # Struct with mixed fields
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.struct(
                            [
                                Field("a", DataType.uint64()),
                                Field("b", DataType.timestamp("ns")),
                                Field("c", DataType.uint32()),
                            ]
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.struct(
                            [
                                Field("a", DataType.int64()),
                                Field("b", DataType.timestamp("us")),
                                Field("c", DataType.int32()),
                            ]
                        ),
                    )
                ]
            ),
        ),
        # Nested struct in list
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.list(
                            DataType.struct(
                                [
                                    Field("a", DataType.uint8(), nullable=False),
                                    Field("b", DataType.timestamp("s")),
                                ]
                            )
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.list(
                            DataType.struct(
                                [
                                    Field("a", DataType.int8(), nullable=False),
                                    Field("b", DataType.timestamp("us")),
                                ]
                            )
                        ),
                    )
                ]
            ),
        ),
        # Mixed schema with multiple fields
        (
            Schema(
                fields=[
                    Field("a", DataType.uint16()),
                    Field("b", DataType.timestamp("ms", tz="Europe/Berlin")),
                    Field(
                        "d",
                        DataType.struct(
                            [
                                Field("x", DataType.uint32()),
                                Field("y", DataType.int64()),
                            ]
                        ),
                    ),
                ]
            ),
            Schema(
                fields=[
                    Field("a", DataType.int16()),
                    Field("b", DataType.timestamp("us", tz="UTC")),
                    Field(
                        "d",
                        DataType.struct(
                            [Field("x", DataType.int32()), Field("y", DataType.int64())]
                        ),
                    ),
                ]
            ),
        ),
        # Field metadata preservations
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.uint16(),
                        metadata={"description": "an unsigned int"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.int16(),
                        metadata={"description": "an unsigned int"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "bar",
                        DataType.timestamp("ns"),
                        nullable=True,
                        metadata={"origin": "sensor_7"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "bar",
                        DataType.timestamp("us"),
                        nullable=True,
                        metadata={"origin": "sensor_7"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "record",
                        DataType.struct(
                            [
                                Field(
                                    "id",
                                    DataType.uint32(),
                                    metadata={"index": "primary"},
                                ),
                                Field(
                                    "value",
                                    DataType.timestamp("s"),
                                    nullable=True,
                                    metadata={"unit": "seconds"},
                                ),
                            ]
                        ),
                        metadata={"type": "event"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "record",
                        DataType.struct(
                            [
                                Field(
                                    "id",
                                    DataType.int32(),
                                    metadata={"index": "primary"},
                                ),
                                Field(
                                    "value",
                                    DataType.timestamp("us"),
                                    nullable=True,
                                    metadata={"unit": "seconds"},
                                ),
                            ]
                        ),
                        metadata={"type": "event"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "sensor_readings",
                        DataType.list(
                            Field(
                                "reading",
                                DataType.uint8(),
                                nullable=False,
                                metadata={"unit": "celsius"},
                            )
                        ),
                        metadata={"shape": "1D"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "sensor_readings",
                        DataType.list(
                            Field(
                                "reading",
                                DataType.int8(),
                                nullable=False,
                                metadata={"unit": "celsius"},
                            )
                        ),
                        metadata={"shape": "1D"},
                    )
                ]
            ),
        ),
    ],
)
def test_schema_conversion(input_schema: Schema, expected_schema: Schema):
    assert expected_schema == _convert_arro3_schema_to_delta(input_schema)


@pytest.mark.pandas
def test_merge_casting_table_provider(tmp_path):
    import pandas as pd

    df = pd.DataFrame(
        {
            "a": 1,
            "ts": pd.date_range(
                "2021-01-01", "2021-01-02", freq="h", tz="America/Chicago"
            ),
        }
    )
    write_deltalake(tmp_path, df, mode="overwrite")

    df2 = pd.DataFrame(
        {
            "a": 2,
            "ts": pd.date_range(
                "2021-01-01", "2021-01-03", freq="h", tz="America/Chicago"
            ),
        }
    )

    dt = DeltaTable(tmp_path)
    dt.merge(
        df2,
        predicate="source.ts = target.ts",
        source_alias="source",
        target_alias="target",
    ).when_matched_update_all().when_not_matched_insert_all().execute()


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_pandas_null_columns_to_existing_table_should_work(tmp_path: pathlib.Path):
    import pandas as pd
    import pyarrow as pa

    initial_data = pa.table(
        {
            "id": [1, 2, 3],
            "name": ["Alice", "Bob", "Charlie"],
            "age": [25, 30, 35],
            "score": [85.5, 92.0, 88.5],
        }
    )

    write_deltalake(tmp_path, initial_data)

    df_with_nulls = pd.DataFrame(
        {
            "id": [4, 5],
            "name": [None, None],  # object dtype -> null type in Arrow
            "age": [None, None],  # object dtype -> null type in Arrow
            "score": [None, None],  # object dtype -> null type in Arrow
        }
    )

    arrow_table = pa.Table.from_pandas(df_with_nulls)

    assert arrow_table.schema.field("name").type == pa.null()
    assert arrow_table.schema.field("age").type == pa.null()
    assert arrow_table.schema.field("score").type == pa.null()

    write_deltalake(tmp_path, df_with_nulls, mode="append")

    updated_dt = DeltaTable(tmp_path)
    result_df = updated_dt.to_pandas()

    # Should have 5 rows total (3 initial + 2 appended)
    assert len(result_df) == 5

    # Check that all expected IDs are present (order may vary)
    all_ids = sorted(result_df["id"].tolist())
    assert all_ids == [1, 2, 3, 4, 5]

    # Check that rows with IDs 4 and 5 have null values for other columns
    new_rows = result_df[result_df["id"].isin([4, 5])]
    assert len(new_rows) == 2
    assert pd.isna(new_rows["name"]).all()
    assert pd.isna(new_rows["age"]).all()
    assert pd.isna(new_rows["score"]).all()


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_prevents_infinite_recursion():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field("id", DataType.int64()),
            Field("problematic_field", DataType.null()),
        ]
    )

    existing_schema = Schema(
        [
            Field("id", DataType.int64()),
            Field("problematic_field", DataType.null()),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema, existing_schema)

    assert converted.field("id").type == DataType.int64()
    assert DataType.is_null(converted.field("problematic_field").type)


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_with_mixed_types():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field("concrete_field", DataType.int64()),
            Field("null_field", DataType.null()),
            Field("missing_field", DataType.string()),
        ]
    )

    existing_schema = Schema(
        [
            Field("concrete_field", DataType.int64()),
            Field("null_field", DataType.string()),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema, existing_schema)

    assert converted.field("concrete_field").type == DataType.int64()
    assert converted.field("null_field").type == DataType.string()
    assert converted.field("missing_field").type == DataType.string()


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_without_existing_schema():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field("id", DataType.int64()),
            Field("null_field", DataType.null()),
            Field("timestamp_field", DataType.timestamp("ns")),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema)

    assert converted.field("id").type == DataType.int64()
    assert DataType.is_null(converted.field("null_field").type)
    assert converted.field("timestamp_field").type == DataType.timestamp("us")


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_field_not_found():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field("missing_field", DataType.null()),
        ]
    )

    existing_schema = Schema(
        [
            Field("other_field", DataType.string()),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema, existing_schema)

    assert DataType.is_null(converted.field("missing_field").type)


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_no_field_name():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field("list_field", DataType.list(Field("element", DataType.null()))),
        ]
    )

    existing_schema = Schema(
        [
            Field("list_field", DataType.list(Field("element", DataType.string()))),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema, existing_schema)

    list_field = converted.field("list_field")
    assert DataType.is_list(list_field.type)


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_null_conversion_with_struct_types():
    from deltalake.writer._conversion import _convert_arro3_schema_to_delta

    source_schema = Schema(
        [
            Field(
                "struct_field",
                DataType.struct(
                    [
                        Field("inner_null", DataType.null()),
                        Field("inner_int", DataType.int32()),
                    ]
                ),
            ),
        ]
    )

    existing_schema = Schema(
        [
            Field(
                "struct_field",
                DataType.struct(
                    [
                        Field("inner_null", DataType.string()),
                        Field("inner_int", DataType.int32()),
                    ]
                ),
            ),
        ]
    )

    converted = _convert_arro3_schema_to_delta(source_schema, existing_schema)

    struct_field = converted.field("struct_field")
    assert DataType.is_struct(struct_field.type)
    inner_fields = struct_field.type.fields
    assert DataType.is_null(inner_fields[0].type)
    assert inner_fields[1].type == DataType.int32()



import atexit
from typing import Optional

from deltalake._internal import (
    TableFeatures,
    Transaction,
    __version__,
    rust_core_version,
)
from deltalake._internal import (
    init_tracing as _init_tracing,
)
from deltalake._internal import (
    shutdown_tracing as _shutdown_tracing,
)
from deltalake.query import QueryBuilder
from deltalake.schema import DataType, Field, Schema
from deltalake.table import DeltaTable, Metadata
from deltalake.transaction import CommitProperties, PostCommitHookProperties
from deltalake.writer import (
    BloomFilterProperties,
    ColumnProperties,
    WriterProperties,
    convert_to_deltalake,
    write_deltalake,
)


def init_tracing(endpoint: Optional[str] = None) -> None:
    """
    Initialize OpenTelemetry tracing for delta-rs operations.

    Args:
        endpoint: The OTLP HTTP endpoint URL. If not provided, uses the
            OTEL_EXPORTER_OTLP_ENDPOINT environment variable or defaults to
            "http://localhost:4318/v1/traces"

    Raises:
        RuntimeError: If tracing initialization fails

    Note:
        - Tracing will be automatically shut down when the Python process exits
        - Use OTEL_EXPORTER_OTLP_HEADERS environment variable for authentication
          (format: "key1=value1,key2=value2")
        - Use RUST_LOG environment variable to control log level filtering
          (e.g., "info", "debug", "deltalake=debug")

    Example:
        Basic usage with defaults:
        ```python
        import deltalake

        deltalake.init_tracing()
        deltalake.write_deltalake("my_table", data)
        dt = deltalake.DeltaTable("my_table")
        ```

        Using custom endpoint:
        ```python
        import deltalake

        deltalake.init_tracing(endpoint="http://localhost:4318/v1/traces")
        ```

        Using environment variables for authentication:
        ```python
        import os
        import deltalake

        os.environ["OTEL_EXPORTER_OTLP_ENDPOINT"] = "https://api.honeycomb.io/v1/traces"
        os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = "x-honeycomb-team=your-api-key"
        os.environ["RUST_LOG"] = "deltalake=debug"
        deltalake.init_tracing()
        ```
    """
    _init_tracing(endpoint)
    atexit.register(_shutdown_tracing)


__all__ = [
    "BloomFilterProperties",
    "ColumnProperties",
    "CommitProperties",
    "DataType",
    "DeltaTable",
    "Field",
    "Metadata",
    "PostCommitHookProperties",
    "QueryBuilder",
    "Schema",
    "TableFeatures",
    "Transaction",
    "WriterProperties",
    "__version__",
    "convert_to_deltalake",
    "init_tracing",
    "rust_core_version",
    "write_deltalake",
]



#!/usr/bin/env python3

import os
import pathlib
import shutil

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import DeltaTable, write_deltalake


def clean_data_dir(data_path):
    if os.path.exists(data_path):
        try:
            shutil.rmtree(data_path)
        except Exception as e:
            print(f"Error deleting directory {data_path}: {e}")


def print_log_dir(path):
    # Show log contents
    log_path = os.path.join(path, "_delta_log")
    if os.path.exists(log_path):
        print("Delta log contents:")
        for file in sorted(os.listdir(log_path)):
            print(f"  {file}")


def valid_gc_data(version) -> Table:
    id_col = ArrowField("id", DataType.int32(), nullable=True)
    gc = ArrowField("gc", DataType.int32(), nullable=True).with_metadata(
        {"delta.generationExpression": "10"}
    )
    data = Table.from_pydict(
        {"id": Array([version, version], type=id_col), "gc": Array([10, 10], type=gc)},
    )
    return data


@pytest.mark.pandas
def test_failed_cleanup(tmp_path: pathlib.Path):
    data_path = tmp_path
    clean_data_dir(data_path)

    # write 10 versions of the data
    for i in range(10):
        data = valid_gc_data(i)
        write_deltalake(
            data_path,
            mode="overwrite",
            data=data,
            configuration={
                "delta.minWriterVersion": "7",
                "delta.logRetentionDuration": "interval 0 day",
            },
        )

    # checkpoint final version
    table = DeltaTable(data_path)
    table.create_checkpoint()

    # show log contents
    print("log contents before cleanup:")
    print_log_dir(data_path)

    # Call cleanup metadata
    table = DeltaTable(data_path, version=5)
    # table.create_checkpoint()  # Workaround, manually create checkpoint needed to load version >= 5
    table.cleanup_metadata()

    # show log contents
    print("\n################################################")
    print("log contents after cleanup:")
    print_log_dir(data_path)

    # Load old version
    table = DeltaTable(data_path, version=5)
    df2 = table.to_pandas()
    print("\n################################################")
    print("Version 5 of the data:")
    print(df2)



import pathlib
from typing import TYPE_CHECKING

import pytest
from arro3.core import Array, DataType, Field, Schema, Table

from deltalake import CommitProperties, DeltaTable, TableFeatures, write_deltalake
from deltalake.exceptions import DeltaError, DeltaProtocolError
from deltalake.schema import Field as DeltaField
from deltalake.schema import PrimitiveType, StructType

if TYPE_CHECKING:
    pass


def test_add_constraint(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": "price >= 0"})

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "ADD CONSTRAINT"
    assert dt.version() == 1
    assert dt.metadata().configuration == {
        "delta.constraints.check_price": "price >= 0"
    }
    assert dt.protocol().min_writer_version == 3

    with pytest.raises(DeltaError):
        # Invalid constraint
        dt.alter.add_constraint({"check_price": "price < 0"})

    with pytest.raises(DeltaProtocolError):
        data = Table(
            {
                "id": Array(["1"], DataType.string()),
                "price": Array([-1], DataType.int64()),
                "sold": Array(list(range(1)), DataType.int32()),
                "deleted": Array([False] * 1, DataType.bool()),
            },
            schema=Schema(
                fields=[
                    Field("id", type=DataType.string(), nullable=True),
                    Field("price", type=DataType.int64(), nullable=True),
                    Field("sold", type=DataType.int32(), nullable=True),
                    Field("deleted", type=DataType.bool(), nullable=True),
                ]
            ),
        )

        write_deltalake(tmp_path, data, mode="append")


def test_add_constraint_roundtrip_metadata(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.alter.add_constraint(
        {"check_price2": "price >= 0"}, commit_properties=commit_properties
    )

    assert dt.history(1)[0]["userName"] == "John Doe"


def test_drop_constraint(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": "price >= 0"})
    assert dt.protocol().min_writer_version == 3
    dt.alter.drop_constraint(name="check_price")
    last_action = dt.history(1)[0]
    assert last_action["operation"] == "DROP CONSTRAINT"
    assert dt.version() == 2
    assert dt.metadata().configuration == {}
    assert dt.protocol().min_writer_version == 3


def test_drop_constraint_invalid(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": "price >= 0"})
    with pytest.raises(DeltaError):
        dt.alter.drop_constraint(name="invalid_constraint_name")

    assert dt.metadata().configuration == {
        "delta.constraints.check_price": "price >= 0"
    }
    assert dt.protocol().min_writer_version == 3


def test_drop_constraint_invalid_ignore(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": "price >= 0"})
    dt.alter.drop_constraint(name="invalid_constraint_name", raise_if_not_exists=False)


def test_drop_constraint_roundtrip_metadata(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price2": "price >= 0"})
    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.alter.drop_constraint("check_price2", commit_properties=commit_properties)

    assert dt.history(1)[0]["userName"] == "John Doe"


@pytest.mark.parametrize("min_writer_version", ["2", "3", "4", "5", "6", "7"])
def test_set_table_properties_min_writer_version(
    tmp_path: pathlib.Path,
    sample_table: Table,
    min_writer_version: str,
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)

    configuration = {"delta.minWriterVersion": min_writer_version}
    dt.alter.set_table_properties(configuration)

    protocol = dt.protocol()

    assert dt.metadata().configuration == configuration
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == int(min_writer_version)


def test_set_table_properties_invalid_min_writer_version(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    with pytest.raises(DeltaError):
        dt.alter.set_table_properties({"delta.minWriterVersion": "8"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {}
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 2


@pytest.mark.parametrize("min_reader_version", ["1", "2", "3"])
def test_set_table_properties_min_reader_version(
    tmp_path: pathlib.Path,
    sample_table: Table,
    min_reader_version: str,
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    configuration = {
        "delta.minReaderVersion": min_reader_version,
        "delta.minWriterVersion": "7",
    }
    dt.alter.set_table_properties(configuration)

    protocol = dt.protocol()
    assert dt.metadata().configuration == configuration
    assert protocol.min_reader_version == int(min_reader_version)
    assert protocol.min_writer_version == 7


def test_set_table_properties_invalid_min_reader_version(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    with pytest.raises(DeltaError):
        dt.alter.set_table_properties({"delta.minReaderVersion": "8"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {}
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 2


def test_set_table_properties_enable_cdf(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    dt.alter.set_table_properties({"delta.enableChangeDataFeed": "true"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {"delta.enableChangeDataFeed": "true"}
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 4


def test_set_table_properties_enable_cdf_invalid(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    with pytest.raises(DeltaError):
        dt.alter.set_table_properties({"delta.enableChangeDataFeed": "wrong"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {}
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 2


def test_set_table_properties_enable_cdf_value_false(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    dt.alter.set_table_properties({"delta.enableChangeDataFeed": "false"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {"delta.enableChangeDataFeed": "false"}
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 2


def test_set_table_properties_enable_cdf_with_writer_version_bumped(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    dt.alter.set_table_properties(
        {"delta.enableChangeDataFeed": "true", "delta.minWriterVersion": "7"}
    )

    protocol = dt.protocol()
    assert dt.metadata().configuration == {
        "delta.enableChangeDataFeed": "true",
        "delta.minWriterVersion": "7",
    }
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 7
    assert protocol.writer_features == ["changeDataFeed"]


def test_set_table_properties_enable_cdf_and_deletion_vectors(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    dt.alter.set_table_properties(
        {"delta.enableChangeDataFeed": "true", "delta.enableDeletionVectors": "true"}
    )

    protocol = dt.protocol()
    assert dt.metadata().configuration == {
        "delta.enableChangeDataFeed": "true",
        "delta.enableDeletionVectors": "true",
    }
    assert protocol.min_reader_version == 3
    assert protocol.min_writer_version == 7
    assert list(sorted(protocol.writer_features)) == [  # type: ignore
        "changeDataFeed",
        "deletionVectors",
    ]
    assert protocol.reader_features == ["deletionVectors"]


def test_convert_checkConstraints_to_feature_after_version_upgrade(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": "price >= 0"})

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "ADD CONSTRAINT"
    assert dt.version() == 1
    assert dt.metadata().configuration == {
        "delta.constraints.check_price": "price >= 0"
    }
    assert dt.protocol().min_writer_version == 3

    dt.alter.set_table_properties({"delta.enableDeletionVectors": "true"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {
        "delta.constraints.check_price": "price >= 0",
        "delta.enableDeletionVectors": "true",
    }
    assert protocol.min_reader_version == 3
    assert protocol.min_writer_version == 7
    assert list(sorted(protocol.writer_features)) == [  # type: ignore
        "checkConstraints",
        "deletionVectors",
    ]
    assert protocol.reader_features == ["deletionVectors"]


def test_set_table_properties_enable_dv(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    dt.alter.set_table_properties({"delta.enableDeletionVectors": "true"})

    protocol = dt.protocol()
    assert dt.metadata().configuration == {"delta.enableDeletionVectors": "true"}
    assert protocol.min_reader_version == 3
    assert protocol.min_writer_version == 7
    assert protocol.writer_features == ["deletionVectors"]
    assert protocol.reader_features == ["deletionVectors"]


def _sort_fields(fields: list[Field]) -> list[DeltaField]:
    return list(sorted(iter(fields), key=lambda x: (x.name, str(x.type))))


def test_add_column_primitive(existing_sample_table: DeltaTable):
    current_fields = existing_sample_table.schema().fields

    new_fields_to_add = [
        DeltaField("foo", PrimitiveType("integer")),
        DeltaField("bar", PrimitiveType("float")),
    ]

    existing_sample_table.alter.add_columns(new_fields_to_add)
    new_fields = existing_sample_table.schema().fields

    assert _sort_fields(new_fields) == _sort_fields(
        [*current_fields, *new_fields_to_add]
    )


@pytest.mark.pyarrow
def test_add_field_in_struct_column(existing_table: DeltaTable):
    current_fields = existing_table.schema().fields

    new_fields_to_add = [
        DeltaField("struct", StructType([DeltaField("z", PrimitiveType("float"))])),
    ]

    existing_table.alter.add_columns(new_fields_to_add)
    new_fields = existing_table.schema().fields

    new_field = DeltaField(
        "struct",
        StructType(
            [
                DeltaField("x", PrimitiveType("long")),
                DeltaField("y", PrimitiveType("string")),
                DeltaField("z", PrimitiveType("float")),
            ]
        ),
    )
    assert _sort_fields(new_fields) == _sort_fields(
        [*[field for field in current_fields if field.name != "struct"], new_field]
    )


def test_add_timestamp_ntz_column(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    dt = DeltaTable(tmp_path)
    current_fields = dt.schema().fields

    new_fields_to_add = DeltaField("timestamp_ntz_col", PrimitiveType("timestamp_ntz"))

    dt.alter.add_columns(new_fields_to_add)
    new_fields = dt.schema().fields
    new_protocol = dt.protocol()

    assert _sort_fields(new_fields) == _sort_fields(
        [*current_fields, new_fields_to_add]
    )
    assert new_protocol.min_reader_version == 3
    assert new_protocol.min_writer_version == 7
    assert new_protocol.reader_features == ["timestampNtz"]
    assert new_protocol.writer_features == ["timestampNtz"]


features = [
    TableFeatures.ChangeDataFeed,
    TableFeatures.DeletionVectors,
    TableFeatures.ColumnMapping,
    TableFeatures.TimestampWithoutTimezone,
    TableFeatures.V2Checkpoint,
    TableFeatures.AppendOnly,
    TableFeatures.AppendOnly,
    TableFeatures.Invariants,
    TableFeatures.CheckConstraints,
    TableFeatures.GeneratedColumns,
    TableFeatures.IdentityColumns,
    TableFeatures.RowTracking,
    TableFeatures.DomainMetadata,
    TableFeatures.IcebergCompatV1,
]

all_features = []
all_features.extend(features)
all_features.append(features)


@pytest.mark.pyarrow
@pytest.mark.parametrize("feature", all_features)
def test_add_feature_variations(existing_table: DeltaTable, feature):
    """Existing table already has timestampNtz so it's already at v3,7"""
    existing_table.alter.add_feature(
        feature=feature,
        allow_protocol_versions_increase=False,
    )
    last_action = existing_table.history(1)[0]
    assert last_action["operation"] == "ADD FEATURE"
    assert existing_table.version() == 1


def test_add_features_disallowed_protocol_increase(existing_sample_table: DeltaTable):
    with pytest.raises(
        DeltaError,
        match="Generic DeltaTable error: Table feature enables writer feature, but min_writer is not v7. Set allow_protocol_versions_increase or increase version explicitly through set_tbl_properties",
    ):
        existing_sample_table.alter.add_feature(
            feature=TableFeatures.ChangeDataFeed,
            allow_protocol_versions_increase=False,
        )
    with pytest.raises(
        DeltaError,
        match="Generic DeltaTable error: Table feature enables reader and writer feature, but reader is not v3, and writer not v7. Set allow_protocol_versions_increase or increase versions explicitly through set_tbl_properties",
    ):
        existing_sample_table.alter.add_feature(
            feature=TableFeatures.DeletionVectors,
            allow_protocol_versions_increase=False,
        )


def test_add_features(existing_sample_table: DeltaTable):
    existing_sample_table.alter.add_feature(
        feature=features,
        allow_protocol_versions_increase=True,
    )
    protocol = existing_sample_table.protocol()

    assert sorted(protocol.reader_features) == sorted(  # type: ignore
        ["v2Checkpoint", "columnMapping", "deletionVectors", "timestampNtz"]
    )
    assert sorted(protocol.writer_features) == sorted(  # type: ignore
        [
            "appendOnly",
            "changeDataFeed",
            "checkConstraints",
            "columnMapping",
            "deletionVectors",
            "domainMetadata",
            "generatedColumns",
            "icebergCompatV1",
            "identityColumns",
            "invariants",
            "rowTracking",
            "timestampNtz",
            "v2Checkpoint",
        ]
    )  # type: ignore


def test_set_column_metadata(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.set_column_metadata("price", {"comment": "my comment"})

    fields_by_name = {field.name: field for field in dt.schema().fields}
    assert fields_by_name["price"].metadata == {"comment": "my comment"}

    with pytest.raises(DeltaError):
        # Can't set metadata for non existing column.
        dt.alter.set_column_metadata("non_existing_column", {"comment": "my comment"})


def test_set_table_name(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    initial_metadata = dt.metadata()
    assert initial_metadata.name is None

    dt.alter.set_table_name("my_awesome_table")
    updated_metadata = dt.metadata()

    assert updated_metadata.name == "my_awesome_table"

    assert dt.version() == 1

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "UPDATE TABLE METADATA"


def test_set_table_description(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)
    initial_metadata = dt.metadata()
    assert initial_metadata.description is None

    dt.alter.set_table_description("A wonderful sample table for testing")
    updated_metadata = dt.metadata()

    assert updated_metadata.description == "A wonderful sample table for testing"

    assert dt.version() == 1

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "UPDATE TABLE METADATA"


def test_set_table_name_overwrite(tmp_path: pathlib.Path, sample_table: Table):
    """Test overwriting an existing table name."""
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)
    dt.alter.set_table_name("initial_name")
    dt.alter.set_table_name("new_name")
    updated_metadata = dt.metadata()

    assert updated_metadata.name == "new_name"


def test_set_table_description_overwrite(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)
    dt.alter.set_table_description("initial description")
    dt.alter.set_table_description("updated description")

    updated_metadata = dt.metadata()

    assert updated_metadata.description == "updated description"

    assert dt.version() == 2


def test_set_table_name_character_limit(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)
    dt = DeltaTable(tmp_path)

    name_255_chars = "x" * 255
    dt.alter.set_table_name(name_255_chars)
    assert dt.metadata().name == name_255_chars

    name_256_chars = "y" * 256
    with pytest.raises(
        DeltaError,
        match="Table metadata is invalid: name: Table name cannot be empty and cannot exceed 255 characters",
    ):
        dt.alter.set_table_name(name_256_chars)


def test_set_table_description_character_limit(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table)
    dt = DeltaTable(tmp_path)

    desc_4000_chars = "x" * 4000
    dt.alter.set_table_description(desc_4000_chars)
    assert dt.metadata().description == desc_4000_chars

    desc_4001_chars = "y" * 4001
    with pytest.raises(
        DeltaError,
        match="Table metadata is invalid: description: Table description cannot exceed 4000 characters",
    ):
        dt.alter.set_table_description(desc_4001_chars)


def test_set_table_metadata_name_none_should_raise_error(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table)
    dt = DeltaTable(tmp_path)

    with pytest.raises(
        DeltaError,
        match="Table metadata is invalid: name: Table name cannot be empty and cannot exceed 255 characters",
    ):
        dt.alter.set_table_name("")



from __future__ import annotations

from arro3.core import DataType
from arro3.core import Schema as Arro3Schema


def _convert_arro3_schema_to_delta(
    schema: Arro3Schema,
    existing_schema: Arro3Schema | None = None,
) -> Arro3Schema:
    """Convert a arro3 schema to a schema compatible with Delta Lake. Converts unsigned to signed equivalent, and
    converts all timestamps to `us` timestamps. Also handles null column types by converting them to match
    corresponding fields in the existing table schema.

    Args:
        schema (Arro3Schema): Source schema.
        existing_schema (Arro3Schema, optional): Existing table schema to match null types against. Defaults to None.

    Returns:
        Arro3Schema: Delta-compatible schema with converted types.
    """

    def dtype_to_delta_dtype(
        dtype: DataType, field_name: str | None = None
    ) -> DataType:
        if (
            DataType.is_null(dtype)
            and existing_schema is not None
            and field_name is not None
        ):
            try:
                existing_field = existing_schema.field(field_name)
                # Prevent infinite recursion: if existing field is also null, keep as null
                if DataType.is_null(existing_field.type):
                    return dtype
                return dtype_to_delta_dtype(existing_field.type, None)
            except (KeyError, IndexError):
                return dtype

        # Handle nested types
        if (
            DataType.is_list(dtype)
            or DataType.is_large_list(dtype)
            or DataType.is_fixed_size_list(dtype)
            or DataType.is_list_view(dtype)
            or DataType.is_large_list_view(dtype)
        ):
            return list_to_delta_dtype(dtype)
        elif DataType.is_struct(dtype):
            return struct_to_delta_dtype(dtype)
        elif DataType.is_timestamp(dtype):
            if dtype.tz is None:
                return DataType.timestamp("us")
            else:
                return DataType.timestamp("us", tz="UTC")
        elif DataType.is_fixed_size_binary(dtype):
            return DataType.binary()
        elif DataType.is_unsigned_integer(dtype):
            if DataType.is_uint16(dtype):
                return DataType.int16()
            elif DataType.is_uint32(dtype):
                return DataType.int32()
            elif DataType.is_uint64(dtype):
                return DataType.int64()
            elif DataType.is_uint8(dtype):
                return DataType.int8()
            else:
                raise NotImplementedError
        else:
            return dtype

    def list_to_delta_dtype(
        dtype: DataType,
    ) -> DataType:
        nested_dtype = dtype.value_type
        inner_field = dtype.value_field

        assert nested_dtype is not None
        assert inner_field is not None

        inner_field_casted = inner_field.with_type(
            dtype_to_delta_dtype(nested_dtype, None)
        )

        if DataType.is_large_list(dtype):
            return DataType.large_list(inner_field_casted)
        elif DataType.is_fixed_size_list(dtype):
            # Fixed sized lists can come in from polars via their Array type.
            # These may carry array field names of "item" rather than "element"
            # which is expected everywhere else. Converting the field name and
            # then passing the field through for further casting in Rust will
            # accommodate this
            #
            # See also: <https://github.com/delta-io/delta-rs/issues/3566>
            if inner_field_casted.name == "item":
                inner_field_casted = inner_field_casted.with_name("element")
            return DataType.list(inner_field_casted, dtype.list_size)
        elif DataType.is_large_list_view(dtype):
            return DataType.large_list_view(inner_field_casted)
        elif DataType.is_list_view(dtype):
            return DataType.list_view(inner_field_casted)
        elif DataType.is_list(dtype):
            return DataType.list(inner_field_casted)
        else:
            raise NotImplementedError

    def struct_to_delta_dtype(dtype: DataType) -> DataType:
        fields_cast = [
            f.with_type(dtype_to_delta_dtype(f.type, f.name)) for f in dtype.fields
        ]
        return DataType.struct(fields_cast)

    return Arro3Schema(
        [f.with_type(dtype_to_delta_dtype(f.type, f.name)) for f in schema]  # type: ignore[attr-defined]
    )



from __future__ import annotations

import json
from collections.abc import Generator, Iterable, Mapping
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    NamedTuple,
    Union,
)

from arro3.core import RecordBatch, RecordBatchReader
from arro3.core.types import (
    ArrowArrayExportable,
    ArrowSchemaExportable,
    ArrowStreamExportable,
)

from deltalake._internal import (
    DeltaError,
    PyMergeBuilder,
    RawDeltaTable,
    TableFeatures,
)
from deltalake._internal import create_deltalake as _create_deltalake
from deltalake._util import encode_partition_value
from deltalake.exceptions import DeltaProtocolError
from deltalake.schema import Field as DeltaField
from deltalake.schema import Schema as DeltaSchema
from deltalake.writer._conversion import _convert_arro3_schema_to_delta

if TYPE_CHECKING:
    import os

    import pandas as pd
    import pyarrow
    import pyarrow.fs as pa_fs
    from pyarrow.dataset import (
        Expression,
        ParquetReadOptions,
    )

    from deltalake.transaction import (
        AddAction,
        CommitProperties,
        PostCommitHookProperties,
    )
    from deltalake.writer.properties import WriterProperties


MAX_SUPPORTED_PYARROW_WRITER_VERSION = 7
NOT_SUPPORTED_PYARROW_WRITER_VERSIONS = [3, 4, 5, 6]
SUPPORTED_WRITER_FEATURES = {"appendOnly", "invariants", "timestampNtz"}

MAX_SUPPORTED_READER_VERSION = 3
NOT_SUPPORTED_READER_VERSION = 2
SUPPORTED_READER_FEATURES = {"timestampNtz"}

FSCK_METRICS_FILES_REMOVED_LABEL = "files_removed"

FilterLiteralType = tuple[str, str, Any]
FilterConjunctionType = list[FilterLiteralType]
FilterDNFType = list[FilterConjunctionType]
FilterType = Union[FilterConjunctionType, FilterDNFType]
PartitionFilterType = list[tuple[str, str, Union[str, list[str]]]]


@dataclass(init=False)
class Metadata:
    """Create a Metadata instance."""

    def __init__(self, table: RawDeltaTable) -> None:
        self._metadata = table.metadata()

    @property
    def id(self) -> int:
        """Return the unique identifier of the DeltaTable."""
        return self._metadata.id

    @property
    def name(self) -> str:
        """Return the user-provided identifier of the DeltaTable."""
        return self._metadata.name

    @property
    def description(self) -> str:
        """Return the user-provided description of the DeltaTable."""
        return self._metadata.description

    @property
    def partition_columns(self) -> list[str]:
        """Return an array containing the names of the partitioned columns of the DeltaTable."""
        return self._metadata.partition_columns

    @property
    def created_time(self) -> int:
        """
        Return The time when this metadata action is created, in milliseconds since the Unix epoch of the DeltaTable.
        """
        return self._metadata.created_time

    @property
    def configuration(self) -> dict[str, str]:
        """Return the DeltaTable properties."""
        return self._metadata.configuration

    def __str__(self) -> str:
        return (
            f"Metadata(id: {self._metadata.id}, name: {self._metadata.name}, "
            f"description: {self._metadata.description}, partition_columns: {self._metadata.partition_columns}, "
            f"created_time: {self.created_time}, configuration: {self._metadata.configuration})"
        )


class DeltaTableConfig(NamedTuple):
    without_files: bool
    log_buffer_size: int


class ProtocolVersions(NamedTuple):
    min_reader_version: int
    min_writer_version: int
    writer_features: list[str] | None
    reader_features: list[str] | None


@dataclass(init=False)
class DeltaTable:
    """Represents a Delta Table"""

    def __init__(
        self,
        table_uri: str | Path | os.PathLike[str],
        version: int | None = None,
        storage_options: dict[str, str] | None = None,
        without_files: bool = False,
        log_buffer_size: int | None = None,
    ) -> None:
        """
        Create the Delta Table from a path with an optional version.
        Multiple StorageBackends are currently supported: AWS S3, Azure Data Lake Storage Gen2, Google Cloud Storage (GCS) and local URI.
        Depending on the storage backend used, you could provide options values using the ``storage_options`` parameter.

        Args:
            table_uri: the path of the DeltaTable
            version: version of the DeltaTable
            storage_options: a dictionary of the options to use for the storage backend
            without_files: If True, will load table without tracking files.
                                Some append-only applications might have no need of tracking any files. So, the
                                DeltaTable will be loaded with a significant memory reduction.
            log_buffer_size: Number of files to buffer when reading the commit log. A positive integer.
                                Setting a value greater than 1 results in concurrent calls to the storage api.
                                This can decrease latency if there are many files in the log since the last checkpoint,
                                but will also increase memory usage. Possible rate limits of the storage backend should
                                also be considered for optimal performance. Defaults to 4 * number of cpus.

        """
        self._storage_options = storage_options
        self._table = RawDeltaTable(
            str(table_uri),
            version=version,
            storage_options=storage_options,
            without_files=without_files,
            log_buffer_size=log_buffer_size,
        )

    @property
    def table_config(self) -> DeltaTableConfig:
        return DeltaTableConfig(*self._table.table_config())

    @staticmethod
    def is_deltatable(
        table_uri: str, storage_options: dict[str, str] | None = None
    ) -> bool:
        """
        Returns True if a Delta Table exists at specified path.
        Returns False otherwise.

        Args:
            table_uri: the path of the DeltaTable
            storage_options: a dictionary of the options to use for the
                storage backend
        """
        return RawDeltaTable.is_deltatable(table_uri, storage_options)

    @classmethod
    def create(
        cls,
        table_uri: str | Path,
        schema: DeltaSchema | ArrowSchemaExportable,
        mode: Literal["error", "append", "overwrite", "ignore"] = "error",
        partition_by: list[str] | str | None = None,
        name: str | None = None,
        description: str | None = None,
        configuration: Mapping[str, str | None] | None = None,
        storage_options: dict[str, str] | None = None,
        commit_properties: CommitProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
        raise_if_key_not_exists: bool = True,
    ) -> DeltaTable:
        """`CREATE` or `CREATE_OR_REPLACE` a delta table given a table_uri.

        Args:
            table_uri: URI of a table
            schema: Table schema
            mode: How to handle existing data. Default is to error if table already exists.
                If 'append', returns not support error if table exists.
                If 'overwrite', will `CREATE_OR_REPLACE` table.
                If 'ignore', will not do anything if table already exists. Defaults to "error".
            partition_by:  List of columns to partition the table by.
            name: User-provided identifier for this table.
            description: User-provided description for this table.
            configuration:  A map containing configuration options for the metadata action.
            storage_options: Options passed to the object store crate.
            commit_properties: properties of the transaction commit. If None, default values are used.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            raise_if_key_not_exists: Whether to raise an error if the configuration uses keys that are not Delta keys

        Returns:
            DeltaTable: created delta table

        Example:
            ```python
            import pyarrow as pa

            from deltalake import DeltaTable

            dt = DeltaTable.create(
                table_uri="my_local_table",
                schema=pa.schema(
                    [pa.field("foo", pa.string()), pa.field("bar", pa.string())]
                ),
                mode="error",
                partition_by="bar",
            )
            ```
        """
        if isinstance(partition_by, str):
            partition_by = [partition_by]

        if isinstance(table_uri, Path):
            table_uri = str(table_uri)

        if not isinstance(schema, DeltaSchema):
            schema = DeltaSchema.from_arrow(schema)

        _create_deltalake(
            table_uri,
            schema,
            partition_by or [],
            mode,
            raise_if_key_not_exists,
            name,
            description,
            configuration,
            storage_options,
            commit_properties,
            post_commithook_properties,
        )

        return cls(table_uri=table_uri, storage_options=storage_options)

    def version(self) -> int:
        """
        Get the version of the DeltaTable.

        Returns:
            The current version of the DeltaTable
        """
        return self._table.version()

    def partitions(
        self,
        partition_filters: list[tuple[str, str, Any]] | None = None,
    ) -> list[dict[str, str]]:
        """
        Returns the partitions as a list of dicts. Example: `[{'month': '1', 'year': '2020', 'day': '1'}, ...]`

        Args:
            partition_filters: The partition filters that will be used for getting the matched partitions, defaults to `None` (no filtering).
        """

        partitions: list[dict[str, str]] = []
        for partition in self._table.get_active_partitions(partition_filters):
            if not partition:
                continue
            partitions.append({k: v for (k, v) in partition})
        return partitions

    def file_uris(
        self, partition_filters: FilterConjunctionType | None = None
    ) -> list[str]:
        """
        Get the list of files as absolute URIs, including the scheme (e.g. "s3://").

        Local files will be just plain absolute paths, without a scheme. (That is,
        no 'file://' prefix.)

        Use the partition_filters parameter to retrieve a subset of files that match the
        given filters.

        Args:
            partition_filters: the partition filters that will be used for getting the matched files

        Returns:
            list of the .parquet files with an absolute URI referenced for the current version of the DeltaTable

        Predicates are expressed in disjunctive normal form (DNF), like [("x", "=", "a"), ...].
        DNF allows arbitrary boolean logical combinations of single partition predicates.
        The innermost tuples each describe a single partition predicate. The list of inner
        predicates is interpreted as a conjunction (AND), forming a more selective and
        multiple partition predicates. Each tuple has format: (key, op, value) and compares
        the key with the value. The supported op are: `=`, `!=`, `in`, and `not in`. If
        the op is in or not in, the value must be a collection such as a list, a set or a tuple.
        The supported type for value is str. Use empty string `''` for Null partition value.

        Example:
            ```
            ("x", "=", "a")
            ("x", "!=", "a")
            ("y", "in", ["a", "b", "c"])
            ("z", "not in", ["a","b"])
            ```
        """
        return self._table.file_uris(
            self._stringify_partition_values(partition_filters)
        )

    file_uris.__doc__ = ""

    def load_as_version(self, version: int | str | datetime) -> None:
        """
        Load/time travel a DeltaTable to a specified version number, or a timestamp version of the table. If a
        string is passed then the argument should be an RFC 3339 and ISO 8601 date and time string format.
        If a datetime object without a timezone is passed, the UTC timezone will be assumed.

        Args:
            version: the identifier of the version of the DeltaTable to load

        Example:
            **Use a version number**
            ```
            dt = DeltaTable("test_table")
            dt.load_as_version(1)
            ```

            **Use a datetime object**
            ```
            dt.load_as_version(datetime(2023, 1, 1))
            dt.load_as_version(datetime(2023, 1, 1, tzinfo=timezone.utc))
            ```

            **Use a datetime in string format**
            ```
            dt.load_as_version("2018-01-26T18:30:09Z")
            dt.load_as_version("2018-12-19T16:39:57-08:00")
            dt.load_as_version("2018-01-26T18:30:09.453+00:00")
            ```
        """
        if isinstance(version, int):
            self._table.load_version(version)
        elif isinstance(version, datetime):
            if version.tzinfo is None:
                version = version.astimezone(timezone.utc)
            self._table.load_with_datetime(version.isoformat())
        elif isinstance(version, str):
            self._table.load_with_datetime(version)
        else:
            raise TypeError(
                "Invalid datatype provided for version, only int, str or datetime are accepted."
            )

    def load_cdf(
        self,
        starting_version: int = 0,
        ending_version: int | None = None,
        starting_timestamp: str | None = None,
        ending_timestamp: str | None = None,
        columns: list[str] | None = None,
        predicate: str | None = None,
        allow_out_of_range: bool = False,
    ) -> RecordBatchReader:
        """
        Load the Change Data Feed (CDF) from the Delta table as a stream of record batches.

        Parameters:
            starting_version (int): The version of the Delta table to start reading CDF from.
            ending_version (int | None): The version to stop reading CDF at. If None, reads up to the latest version.
            starting_timestamp (str | None): An ISO 8601 timestamp to start reading CDF from. Ignored if starting_version is provided.
            ending_timestamp (str | None): An ISO 8601 timestamp to stop reading CDF at. Ignored if ending_version is provided.
            columns (list[str] | None): A list of column names to include in the output. If None, all columns are included.
            predicate (str | None): An optional SQL predicate to filter the output rows.
            allow_out_of_range (bool): If True, does not raise an error when specified versions or timestamps are outside the table's history.

        Returns:
            RecordBatchReader: An Arrow RecordBatchReader that streams the resulting change data.

        Raises:
            ValueError: If input parameters are invalid or if the specified range is not found (unless allow_out_of_range is True).
        """
        return self._table.load_cdf(
            columns=columns,
            predicate=predicate,
            starting_version=starting_version,
            ending_version=ending_version,
            starting_timestamp=starting_timestamp,
            ending_timestamp=ending_timestamp,
            allow_out_of_range=allow_out_of_range,
        )

    @property
    def table_uri(self) -> str:
        return self._table.table_uri()

    def schema(self) -> DeltaSchema:
        """
        Get the current schema of the DeltaTable.

        Returns:
            the current Schema registered in the transaction log
        """
        return self._table.schema

    def metadata(self) -> Metadata:
        """
        Get the current metadata of the DeltaTable.

        Returns:
            the current Metadata registered in the transaction log
        """
        return Metadata(self._table)

    def protocol(self) -> ProtocolVersions:
        """
        Get the reader and writer protocol versions of the DeltaTable.

        Returns:
            the current ProtocolVersions registered in the transaction log
        """
        return ProtocolVersions(*self._table.protocol_versions())

    def generate(self) -> None:
        """
        Generate symlink manifest for engines that cannot read native Delta Lake tables.

        The generate supports the fairly simple "GENERATE" operation which produces a
        [symlink_format_manifest](https://docs.delta.io/delta-utility/#generate-a-manifest-file) file
        when needed for an external engine such as Presto or BigQuery.

        The "symlink_format_manifest" is not something that has been well documented, but for
        enon-partitioned tables this will generate a `_symlink_format_manifest/manifest` file next to
        the `_delta_log`, for example:

        ```
        COVID-19_NYT
        ├── _delta_log
        │   ├── 00000000000000000000.crc
        │   └── 00000000000000000000.json
        ├── part-00000-a496f40c-e091-413a-85f9-b1b69d4b3b4e-c000.snappy.parquet
        ├── part-00001-9d9d980b-c500-4f0b-bb96-771a515fbccc-c000.snappy.parquet
        ├── part-00002-8826af84-73bd-49a6-a4b9-e39ffed9c15a-c000.snappy.parquet
        ├── part-00003-539aff30-2349-4b0d-9726-c18630c6ad90-c000.snappy.parquet
        ├── part-00004-1bb9c3e3-c5b0-4d60-8420-23261f58a5eb-c000.snappy.parquet
        ├── part-00005-4d47f8ff-94db-4d32-806c-781a1cf123d2-c000.snappy.parquet
        ├── part-00006-d0ec7722-b30c-4e1c-92cd-b4fe8d3bb954-c000.snappy.parquet
        ├── part-00007-4582392f-9fc2-41b0-ba97-a74b3afc8239-c000.snappy.parquet
        └── _symlink_format_manifest
            └── manifest
        ```
        """
        return self._table.generate()

    def history(self, limit: int | None = None) -> list[dict[str, Any]]:
        """
        Run the history command on the DeltaTable.
        The operations are returned in reverse chronological order.

        Args:
            limit: the commit info limit to return

        Returns:
            list of the commit infos registered in the transaction log
        """

        def _backwards_enumerate(
            iterable: list[str], start_end: int
        ) -> Generator[tuple[int, str], None, None]:
            n = start_end
            for elem in iterable:
                yield n, elem
                n -= 1

        commits = list(self._table.history(limit))
        history = []
        for version, commit_info_raw in _backwards_enumerate(
            commits, start_end=self._table.get_latest_version()
        ):
            commit = json.loads(commit_info_raw)
            commit["version"] = version
            history.append(commit)
        return history

    def count(self) -> int:
        """
        Get the approximate row count based on file statistics added to the Delta table.

        This requires that add actions have been added to the Delta table with
        per-file statistics enabled. Because this is an optional field this
        "count" will be less than or equal to the true row count of the table.
        In order to get an exact number of rows a full table scan must happen

        Returns:
            The approximate number of rows for this specific table
        """
        total_rows = 0

        for value in self.get_add_actions().column("num_records").to_pylist():
            # Add action file statistics are optional and so while most modern
            # tables are _likely_ to have this information it is not
            # guaranteed.
            if value is not None:
                total_rows += value

        return total_rows

    def vacuum(
        self,
        retention_hours: int | None = None,
        dry_run: bool = True,
        enforce_retention_duration: bool = True,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
        full: bool = False,
        keep_versions: list[int] | None = None,
    ) -> list[str]:
        """
        Run the Vacuum command on the Delta Table: list and delete files no longer referenced by the Delta table and are older than the retention threshold.

        Args:
            retention_hours: the retention threshold in hours, if none then the value from `delta.deletedFileRetentionDuration` is used or default of 1 week otherwise.
            dry_run: when activated, list only the files, delete otherwise
            enforce_retention_duration: when disabled, accepts retention hours smaller than the value from `delta.deletedFileRetentionDuration`.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.
            full: when set to True, will perform a "full" vacuum and remove all files not referenced in the transaction log
            keep_versions: An optional list of versions to keep. If provided, files from these versions will not be deleted.
        Returns:
            the list of files no longer referenced by the Delta Table and are older than the retention threshold.
        """
        if retention_hours:
            if retention_hours < 0:
                raise ValueError("The retention periods should be positive.")

        return self._table.vacuum(
            dry_run,
            retention_hours,
            enforce_retention_duration,
            commit_properties,
            post_commithook_properties,
            full,
            keep_versions,
        )

    def update(
        self,
        updates: dict[str, str] | None = None,
        new_values: dict[str, int | float | str | datetime | bool | list[Any]]
        | None = None,
        predicate: str | None = None,
        writer_properties: WriterProperties | None = None,
        error_on_type_mismatch: bool = True,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """`UPDATE` records in the Delta Table that matches an optional predicate. Either updates or new_values needs
        to be passed for it to execute.

        Args:
            updates: a mapping of column name to update SQL expression.
            new_values: a mapping of column name to python datatype.
            predicate: a logical expression.
            writer_properties: Pass writer properties to the Rust parquet writer.
            error_on_type_mismatch: specify if update will return error if data types are mismatching :default = True
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.
        Returns:
            the metrics from update

        Example:
            **Update some row values with SQL predicate**

            This is equivalent to `UPDATE table SET deleted = true WHERE id = '3'`
            ```py
            from deltalake import write_deltalake, DeltaTable
            import pandas as pd
            df = pd.DataFrame(
                {"id": ["1", "2", "3"],
                "deleted": [False, False, False],
                "price": [10., 15., 20.]
                })
            write_deltalake("tmp", df)
            dt = DeltaTable("tmp")
            dt.update(predicate="id = '3'", updates = {"deleted": 'True'})

            {'num_added_files': 1, 'num_removed_files': 1, 'num_updated_rows': 1, 'num_copied_rows': 2, 'execution_time_ms': ..., 'scan_time_ms': ...}
            ```

            **Update all row values**

            This is equivalent to ``UPDATE table SET deleted = true, id = concat(id, '_old')``.
            ```py
            dt.update(updates = {"deleted": 'True', "id": "concat(id, '_old')"})

            {'num_added_files': 1, 'num_removed_files': 1, 'num_updated_rows': 3, 'num_copied_rows': 0, 'execution_time_ms': ..., 'scan_time_ms': ...}
            ```

            **Use Python objects instead of SQL strings**

            Use the `new_values` parameter instead of the `updates` parameter. For example,
            this is equivalent to ``UPDATE table SET price = 150.10 WHERE id = '1'``
            ```py
            dt.update(predicate="id = '1_old'", new_values = {"price": 150.10})

            {'num_added_files': 1, 'num_removed_files': 1, 'num_updated_rows': 1, 'num_copied_rows': 2, 'execution_time_ms': ..., 'scan_time_ms': ...}
            ```
        """
        if updates is None and new_values is not None:
            updates = {}
            for key, value in new_values.items():
                if isinstance(value, (int, float, bool, list)):
                    value = str(value)
                elif isinstance(value, str):
                    value = f"'{value}'"
                elif isinstance(value, datetime):
                    value = str(
                        int(value.timestamp() * 1000 * 1000)
                    )  # convert to microseconds
                else:
                    raise TypeError(
                        "Invalid datatype provided in new_values, only int, float, bool, list, str or datetime or accepted."
                    )
                updates[key] = value
        elif updates is not None and new_values is None:
            for key, value in updates.items():
                if not isinstance(value, str) or not isinstance(key, str):
                    raise TypeError(
                        f"The values of the updates parameter must all be SQL strings. Got {updates}. Did you mean to use the new_values parameter?"
                    )

        elif updates is not None and new_values is not None:
            raise ValueError(
                "Passing updates and new_values at same time is not allowed, pick one."
            )
        else:
            raise ValueError(
                "Either updates or new_values need to be passed to update the table."
            )
        metrics = self._table.update(
            updates,
            predicate,
            writer_properties,
            safe_cast=not error_on_type_mismatch,
            commit_properties=commit_properties,
            post_commithook_properties=post_commithook_properties,
        )
        return json.loads(metrics)

    @property
    def optimize(
        self,
    ) -> TableOptimizer:
        """Namespace for all table optimize related methods.

        Returns:
            TableOptimizer: TableOptimizer Object
        """
        return TableOptimizer(self)

    @property
    def alter(
        self,
    ) -> TableAlterer:
        """Namespace for all table alter related methods.

        Returns:
            TableAlterer: TableAlterer Object
        """
        return TableAlterer(self)

    def merge(
        self,
        source: ArrowStreamExportable | ArrowArrayExportable,
        predicate: str,
        source_alias: str | None = None,
        target_alias: str | None = None,
        merge_schema: bool = False,
        error_on_type_mismatch: bool = True,
        writer_properties: WriterProperties | None = None,
        streamed_exec: bool = True,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> TableMerger:
        """Pass the source data which you want to merge on the target delta table, providing a
        predicate in SQL query like format. You can also specify on what to do when the underlying data types do not
        match the underlying table.

        Args:
            source: source data
            predicate: SQL like predicate on how to merge
            source_alias: Alias for the source table
            target_alias: Alias for the target table
            merge_schema: Enable merge schema evolution for mismatch schema between source and target tables
            error_on_type_mismatch: specify if merge will return error if data types are mismatching :default = True
            writer_properties: Pass writer properties to the Rust parquet writer
            streamed_exec: Will execute MERGE using a LazyMemoryExec plan, this improves memory pressure for large source tables. Enabling streamed_exec
                implicitly disables source table stats to derive an early_pruning_predicate
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties for the commit. If None, default values are used.

        Returns:
            TableMerger: TableMerger Object
        """

        source = RecordBatchReader.from_arrow(source)
        compatible_delta_schema = _convert_arro3_schema_to_delta(source.schema)

        py_merge_builder = self._table.create_merge_builder(
            source=source,
            batch_schema=compatible_delta_schema,
            predicate=predicate,
            source_alias=source_alias,
            target_alias=target_alias,
            merge_schema=merge_schema,
            safe_cast=not error_on_type_mismatch,
            streamed_exec=streamed_exec,
            writer_properties=writer_properties,
            commit_properties=commit_properties,
            post_commithook_properties=post_commithook_properties,
        )
        return TableMerger(py_merge_builder, self._table)

    def restore(
        self,
        target: int | datetime | str,
        *,
        ignore_missing_files: bool = False,
        protocol_downgrade_allowed: bool = False,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """
        Run the Restore command on the Delta Table: restore table to a given version or datetime.

        Args:
            target: the expected version will restore, which represented by int, date str or datetime.
            ignore_missing_files: whether the operation carry on when some data files missing.
            protocol_downgrade_allowed: whether the operation when protocol version upgraded.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Returns:
            the metrics from restore.
        """
        if isinstance(target, datetime):
            metrics = self._table.restore(
                target.isoformat(),
                ignore_missing_files=ignore_missing_files,
                protocol_downgrade_allowed=protocol_downgrade_allowed,
                commit_properties=commit_properties,
            )
        else:
            metrics = self._table.restore(
                target,
                ignore_missing_files=ignore_missing_files,
                protocol_downgrade_allowed=protocol_downgrade_allowed,
                commit_properties=commit_properties,
            )
        return json.loads(metrics)

    def to_pyarrow_dataset(
        self,
        partitions: FilterConjunctionType | None = None,
        filesystem: str | pa_fs.FileSystem | None = None,
        parquet_read_options: ParquetReadOptions | None = None,
        schema: pyarrow.Schema | None = None,
        as_large_types: bool = False,
    ) -> "pyarrow.dataset.Dataset":
        """
        Build a PyArrow Dataset using data from the DeltaTable.

        Args:
            partitions: A list of partition filters, see help(DeltaTable.files_by_partitions) for filter syntax
            filesystem: A concrete implementation of the Pyarrow FileSystem or a fsspec-compatible interface. If None, the first file path will be used to determine the right FileSystem
            parquet_read_options: Optional read options for Parquet. Use this to handle INT96 to timestamp conversion for edge cases like 0001-01-01 or 9999-12-31
            schema: The schema to use for the dataset. If None, the schema of the DeltaTable will be used. This can be used to force reading of Parquet/Arrow datatypes
                that DeltaLake can't represent in it's schema (e.g. LargeString).
                If you only need to read the schema with large types (e.g. for compatibility with Polars) you may want to use the `as_large_types` parameter instead.
            as_large_types: get schema with all variable size types (list, binary, string) as large variants (with int64 indices).
                This is for compatibility with systems like Polars that only support the large versions of Arrow types.
                If `schema` is passed it takes precedence over this option.

         More info: https://arrow.apache.org/docs/python/generated/pyarrow.dataset.ParquetReadOptions.html

        Example:
            ``deltalake`` will work with any storage compliant with :class:`pyarrow.fs.FileSystem`, however the root of the filesystem has
            to be adjusted to point at the root of the Delta table. We can achieve this by wrapping the custom filesystem into
            a :class:`pyarrow.fs.SubTreeFileSystem`.
            ```
            import pyarrow.fs as fs
            from deltalake import DeltaTable

            table_uri = "s3://<bucket>/<path>"
            raw_fs, normalized_path = fs.FileSystem.from_uri(table_uri)
            filesystem = fs.SubTreeFileSystem(normalized_path, raw_fs)

            dt = DeltaTable(table_uri)
            ds = dt.to_pyarrow_dataset(filesystem=filesystem)
            ```

        Returns:
            the PyArrow dataset in PyArrow
        """
        try:
            from pyarrow.dataset import (
                FileSystemDataset,
                ParquetFileFormat,
                ParquetFragmentScanOptions,
            )
        except ImportError:
            raise ImportError(
                "Pyarrow is required, install deltalake[pyarrow] for pyarrow read functionality."
            )
        if not self._table.has_files():
            raise DeltaError("Table is instantiated without files.")

        table_protocol = self.protocol()
        if (
            table_protocol.min_reader_version > MAX_SUPPORTED_READER_VERSION
            or table_protocol.min_reader_version == NOT_SUPPORTED_READER_VERSION
        ):
            raise DeltaProtocolError(
                f"The table's minimum reader version is {table_protocol.min_reader_version} "
                f"but deltalake only supports version 1 or {MAX_SUPPORTED_READER_VERSION} "
                f"with these reader features: {SUPPORTED_READER_FEATURES}"
            )
        if (
            table_protocol.min_reader_version >= 3
            and table_protocol.reader_features is not None
        ):
            missing_features = {*table_protocol.reader_features}.difference(
                SUPPORTED_READER_FEATURES
            )
            if len(missing_features) > 0:
                raise DeltaProtocolError(
                    f"The table has set these reader features: {missing_features} "
                    "but these are not yet supported by the deltalake reader."
                )

        if (
            table_protocol.reader_features
            and "columnMapping" in table_protocol.reader_features
        ):
            raise DeltaProtocolError(
                "The table requires reader feature 'columnMapping' "
                "but this is not supported using pyarrow Datasets."
            )

        if (
            table_protocol.reader_features
            and "deletionVectors" in table_protocol.reader_features
        ):
            raise DeltaProtocolError(
                "The table requires reader feature 'deletionVectors' "
                "but this is not supported using pyarrow Datasets."
            )

        import pyarrow
        import pyarrow.fs as pa_fs

        from deltalake.fs import DeltaStorageHandler

        if not filesystem:
            filesystem = pa_fs.PyFileSystem(
                DeltaStorageHandler.from_table(
                    self._table,
                    self._storage_options,
                    self._table.get_add_file_sizes(),
                )
            )

        format = ParquetFileFormat(
            read_options=parquet_read_options,
            default_fragment_scan_options=ParquetFragmentScanOptions(pre_buffer=True),
        )

        if schema is None:
            schema = pyarrow.schema(
                self.schema().to_arrow(as_large_types=as_large_types)
            )

        partitions = self._stringify_partition_values(partitions)

        fragments = [
            format.make_fragment(
                file,
                filesystem=filesystem,
                partition_expression=part_expression,
            )
            for file, part_expression in self._table.dataset_partitions(
                schema, partitions
            )
        ]

        dictionary_columns = format.read_options.dictionary_columns or set()
        if dictionary_columns:
            for index, field in enumerate(schema):
                if field.name in dictionary_columns:
                    dict_field = field.with_type(
                        pyarrow.dictionary(pyarrow.int32(), field.type)
                    )
                    schema = schema.set(index, dict_field)

        return FileSystemDataset(fragments, schema, format, filesystem)

    def to_pyarrow_table(
        self,
        partitions: list[tuple[str, str, Any]] | None = None,
        columns: list[str] | None = None,
        filesystem: str | pa_fs.FileSystem | None = None,
        filters: FilterType | Expression | None = None,
    ) -> "pyarrow.Table":
        """
        Build a PyArrow Table using data from the DeltaTable.

        Args:
            partitions: A list of partition filters, see help(DeltaTable.files_by_partitions) for filter syntax
            columns: The columns to project. This can be a list of column names to include (order and duplicates will be preserved)
            filesystem: A concrete implementation of the Pyarrow FileSystem or a fsspec-compatible interface. If None, the first file path will be used to determine the right FileSystem
            filters: A disjunctive normal form (DNF) predicate for filtering rows, or directly a pyarrow.dataset.Expression. If you pass a filter you do not need to pass ``partitions``
        """
        try:
            from pyarrow.parquet import filters_to_expression  # pyarrow >= 10.0.0
        except ImportError:
            raise ImportError(
                "Pyarrow is required, install deltalake[pyarrow] for pyarrow read functionality."
            )

        if filters is not None:
            filters = filters_to_expression(filters)
        return self.to_pyarrow_dataset(
            partitions=partitions, filesystem=filesystem
        ).to_table(columns=columns, filter=filters)

    def to_pandas(
        self,
        partitions: list[tuple[str, str, Any]] | None = None,
        columns: list[str] | None = None,
        filesystem: str | pa_fs.FileSystem | None = None,
        filters: FilterType | Expression | None = None,
        types_mapper: Callable[[pyarrow.DataType], Any] | None = None,
    ) -> "pd.DataFrame":
        """
        Build a pandas dataframe using data from the DeltaTable.

        Args:
            partitions: A list of partition filters, see help(DeltaTable.files_by_partitions) for filter syntax
            columns: The columns to project. This can be a list of column names to include (order and duplicates will be preserved)
            filesystem: A concrete implementation of the Pyarrow FileSystem or a fsspec-compatible interface. If None, the first file path will be used to determine the right FileSystem
            filters: A disjunctive normal form (DNF) predicate for filtering rows, or directly a pyarrow.dataset.Expression. If you pass a filter you do not need to pass ``partitions``
        """
        return self.to_pyarrow_table(
            partitions=partitions,
            columns=columns,
            filesystem=filesystem,
            filters=filters,
        ).to_pandas(types_mapper=types_mapper)

    def update_incremental(self) -> None:
        """
        Updates the DeltaTable to the latest version by incrementally applying
        newer versions.
        """
        self._table.update_incremental()

    def create_checkpoint(self) -> None:
        """
        Create a checkpoint at the current table version.
        """
        self._table.create_checkpoint()

    def cleanup_metadata(self) -> None:
        """
        Delete expired log files before current version from table. The table log retention is based on
        the `delta.logRetentionDuration` value, 30 days by default.
        """
        self._table.cleanup_metadata()

    def _stringify_partition_values(
        self, partition_filters: FilterConjunctionType | None
    ) -> PartitionFilterType | None:
        if partition_filters is None:
            return partition_filters
        out = []
        for field, op, value in partition_filters:
            str_value: str | list[str]
            if isinstance(value, (list, tuple)):
                str_value = [encode_partition_value(val) for val in value]
            else:
                str_value = encode_partition_value(value)
            out.append((field, op, str_value))
        return out

    def get_add_actions(self, flatten: bool = False) -> RecordBatch:
        """
        Return a dataframe with all current add actions.

        Add actions represent the files that currently make up the table. This
        data is a low-level representation parsed from the transaction log.

        Args:
            flatten: whether to flatten the schema. Partition values columns are
                        given the prefix `partition.`, statistics (null_count, min, and max) are
                        given the prefix `null_count.`, `min.`, and `max.`, and tags the
                        prefix `tags.`. Nested field names are concatenated with `.`.

        Returns:
            a PyArrow RecordBatch containing the add action data.

        Example:
            ```python
            from pprint import pprint
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa
            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data, partition_by=["x"])
            dt = DeltaTable("tmp")
            df = dt.get_add_actions().to_pandas()
            df["path"].sort_values(ignore_index=True)
            0    x=1/0
            1    x=2/0
            2    x=3/0
            ```

            ```python
            df = dt.get_add_actions(flatten=True).to_pandas()
            df["partition.x"].sort_values(ignore_index=True)
            0    1
            1    2
            2    3
            ```
        """
        return self._table.get_add_actions(flatten)

    def delete(
        self,
        predicate: str | None = None,
        writer_properties: WriterProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """Delete records from a Delta Table that satisfy a predicate.

        When a predicate is not provided then all records are deleted from the Delta
        Table. Otherwise a scan of the Delta table is performed to mark any files
        that contain records that satisfy the predicate. Once files are determined
        they are rewritten without the records.

        Args:
            predicate: a SQL where clause. If not passed, will delete all rows.
            writer_properties: Pass writer properties to the Rust parquet writer.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Returns:
            the metrics from delete.
        """
        metrics = self._table.delete(
            predicate,
            writer_properties,
            commit_properties,
            post_commithook_properties,
        )
        return json.loads(metrics)

    def repair(
        self,
        dry_run: bool = False,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """Repair the Delta Table by auditing active files that do not exist in the underlying
        filesystem and removes them. This can be useful when there are accidental deletions or corrupted files.

        Active files are ones that have an add action in the log, but no corresponding remove action.
        This operation creates a new FSCK transaction containing a remove action for each of the missing
        or corrupted files.

        Args:
            dry_run: when activated, list only the files, otherwise add remove actions to transaction log. Defaults to False.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Returns:
            The metrics from repair (FSCK) action.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable('TEST')
            dt.repair(dry_run=False)
            ```
            Results in
            ```
            {'dry_run': False, 'files_removed': ['6-0d084325-6885-4847-b008-82c1cf30674c-0.parquet', 5-4fba1d3e-3e20-4de1-933d-a8e13ac59f53-0.parquet']}
            ```
        """
        metrics = self._table.repair(
            dry_run,
            commit_properties,
            post_commithook_properties,
        )
        deserialized_metrics = json.loads(metrics)
        deserialized_metrics[FSCK_METRICS_FILES_REMOVED_LABEL] = json.loads(
            deserialized_metrics[FSCK_METRICS_FILES_REMOVED_LABEL]
        )
        return deserialized_metrics

    def transaction_version(self, app_id: str) -> int | None:
        """
        Retrieve the latest transaction versions for the given application ID.

        Args:
            app_id (str): The application ID for which to retrieve the latest transaction version.

        Returns:
            int | None: The latest transaction version for the given application ID if it exists, otherwise None.
        """
        return self._table.transaction_version(app_id)

    def create_write_transaction(
        self,
        actions: list[AddAction],
        mode: str,
        schema: DeltaSchema | ArrowSchemaExportable,
        partition_by: list[str] | str | None = None,
        partition_filters: FilterType | None = None,
        commit_properties: CommitProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
    ) -> None:
        if isinstance(partition_by, str):
            partition_by = [partition_by]

        if not isinstance(schema, DeltaSchema):
            schema = DeltaSchema.from_arrow(schema)

        self._table.create_write_transaction(
            actions,
            mode,
            partition_by or [],
            schema,
            partition_filters,
            commit_properties=commit_properties,
            post_commithook_properties=post_commithook_properties,
        )

    def __datafusion_table_provider__(self) -> Any:
        """Return the DataFusion table provider PyCapsule interface.

        To support DataFusion features such as push down filtering, this function will return a PyCapsule
        interface that conforms to the FFI Table Provider required by DataFusion. From an end user perspective
        you should not need to call this function directly. Instead you can use ``register_table_provider`` in
        the DataFusion SessionContext.

        Returns:
            A PyCapsule DataFusion TableProvider interface.

        Example:
            ```python
            from deltalake import DeltaTable, write_deltalake
            from datafusion import SessionContext
            import pyarrow as pa
            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            ctx = SessionContext()
            ctx.register_table_provider("test", table)
            ctx.table("test").show()
            ```
            Results in
            ```
            DataFrame()
            +----+----+----+
            | c3 | c1 | c2 |
            +----+----+----+
            | 4  | 6  | a  |
            | 6  | 5  | b  |
            | 5  | 4  | c  |
            +----+----+----+
            ```
        """
        return self._table.__datafusion_table_provider__()


class TableMerger:
    """API for various table `MERGE` commands."""

    def __init__(
        self,
        builder: PyMergeBuilder,
        table: RawDeltaTable,
    ) -> None:
        self._builder = builder
        self._table = table

    def when_matched_update(
        self, updates: dict[str, str], predicate: str | None = None
    ) -> TableMerger:
        """Update a matched table row based on the rules defined by ``updates``.
        If a ``predicate`` is specified, then it must evaluate to true for the row to be updated.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            updates: a mapping of column name to update SQL expression.
            predicate:  SQL like predicate on when to update.

        Returns:
            TableMerger: TableMerger Object

        Example:
            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "1y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [1], "1y": [7]})

            (
                 dt.merge(
                     source=new_data,
                     predicate="target.x = source.x",
                     source_alias="source",
                     target_alias="target")
                 .when_matched_update(updates={"x": "source.x", "`1y`": "source.`1y`"})
                 .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 1, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 2, 'num_output_rows': 3, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas()
               x  y
            0  1  7
            1  2  5
            2  3  6
            ```
        """
        self._builder.when_matched_update(updates, predicate)
        return self

    def when_matched_update_all(
        self, predicate: str | None = None, except_cols: list[str] | None = None
    ) -> TableMerger:
        """Updating all source fields to target fields, source and target are required to have the same field names.
        If a ``predicate`` is specified, then it must evaluate to true for the row to be updated.
        If ``except_cols`` is specified, then the columns in the exclude list will not be updated.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            predicate: SQL like predicate on when to update all columns.
            except_cols: List of columns to exclude from update.
        Returns:
            TableMerger: TableMerger Object

        Example:
            ** Update all columns **

            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [1], "y": [7]})

            (
                dt.merge(
                    source=new_data,
                    predicate="target.x = source.x",
                    source_alias="source",
                    target_alias="target")
                .when_matched_update_all()
                .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 1, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 2, 'num_output_rows': 3, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas()
               x  y
            0  1  7
            1  2  5
            2  3  6
            ```

            ** Update all columns except `bar` **

            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"foo": [1, 2, 3], "bar": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"foo": [1], "bar": [7]})

            (
                dt.merge(
                    source=new_data,
                    predicate="target.foo = source.foo",
                    source_alias="source",
                    target_alias="target")
                .when_matched_update_all(except_cols=["bar"])
                .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 1, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 2, 'num_output_rows': 3, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas()
               foo  bar
            0  1    4
            1  2    5
            2  3    6
            ```
        """
        maybe_source_alias = self._builder.source_alias
        maybe_target_alias = self._builder.target_alias

        src_alias = (maybe_source_alias + ".") if maybe_source_alias is not None else ""
        trgt_alias = (
            (maybe_target_alias + ".") if maybe_target_alias is not None else ""
        )

        except_columns = except_cols or []

        updates = {
            f"{trgt_alias}`{col.name}`": f"{src_alias}`{col.name}`"
            for col in self._builder.arrow_schema  # type: ignore[attr-defined]
            if col.name not in except_columns
        }

        self._builder.when_matched_update(updates, predicate)
        return self

    def when_matched_delete(self, predicate: str | None = None) -> TableMerger:
        """Delete a matched row from the table only if the given ``predicate`` (if specified) is
        true for the matched row. If not specified it deletes all matches.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
           predicate (str | None, Optional):  SQL like predicate on when to delete.

        Returns:
            TableMerger: TableMerger Object

        Example:
            **Delete on a predicate**

            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [2, 3], "deleted": [False, True]})

            (
                dt.merge(
                    source=new_data,
                    predicate='target.x = source.x',
                    source_alias='source',
                    target_alias='target')
                .when_matched_delete(
                    predicate="source.deleted = true")
                .execute()
            )
            {'num_source_rows': 2, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 0, 'num_target_rows_deleted': 1, 'num_target_rows_copied': 2, 'num_output_rows': 2, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas().sort_values("x", ignore_index=True)
               x  y
            0  1  4
            1  2  5
            ```

            **Delete all records that were matched**
            ```python
            dt = DeltaTable("tmp")
            (
                dt.merge(
                    source=new_data,
                    predicate='target.x = source.x',
                    source_alias='source',
                    target_alias='target')
                .when_matched_delete()
                .execute()
            )
            {'num_source_rows': 2, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 0, 'num_target_rows_deleted': 1, 'num_target_rows_copied': 1, 'num_output_rows': 1, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas()
               x  y
            0  1  4
            ```
        """
        self._builder.when_matched_delete(predicate)
        return self

    def when_not_matched_insert(
        self, updates: dict[str, str], predicate: str | None = None
    ) -> TableMerger:
        """Insert a new row to the target table based on the rules defined by ``updates``. If a
        ``predicate`` is specified, then it must evaluate to true for the new row to be inserted.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            updates (dict):  a mapping of column name to insert SQL expression.
            predicate (str | None, Optional): SQL like predicate on when to insert.

        Returns:
            TableMerger: TableMerger Object

        Example:
            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [4], "y": [7]})

            (
                dt.merge(
                    source=new_data,
                    predicate="target.x = source.x",
                    source_alias="source",
                    target_alias="target",)
                .when_not_matched_insert(
                    updates={
                        "x": "source.x",
                        "y": "source.y",
                    })
                .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 1, 'num_target_rows_updated': 0, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 3, 'num_output_rows': 4, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas().sort_values("x", ignore_index=True)
               x  y
            0  1  4
            1  2  5
            2  3  6
            3  4  7
            ```
        """
        self._builder.when_not_matched_insert(updates, predicate)
        return self

    def when_not_matched_insert_all(
        self, predicate: str | None = None, except_cols: list[str] | None = None
    ) -> TableMerger:
        """Insert a new row to the target table, updating all source fields to target fields. Source and target are
        required to have the same field names. If a ``predicate`` is specified, then it must evaluate to true for
        the new row to be inserted. If ``except_cols`` is specified, then the columns in the exclude list will not be inserted.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            predicate: SQL like predicate on when to insert.
            except_cols: List of columns to exclude from insert.

        Returns:
            TableMerger: TableMerger Object

        Example:
            ** Insert all columns **

            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [4], "y": [7]})

            (
               dt.merge(
                   source=new_data,
                   predicate='target.x = source.x',
                   source_alias='source',
                   target_alias='target')
               .when_not_matched_insert_all()
               .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 1, 'num_target_rows_updated': 0, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 3, 'num_output_rows': 4, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas().sort_values("x", ignore_index=True)
               x  y
            0  1  4
            1  2  5
            2  3  6
            3  4  7
            ```

            ** Insert all columns except `bar` **

            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"foo": [1, 2, 3], "bar": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"foo": [4], "bar": [7]})

            (
               dt.merge(
                   source=new_data,
                   predicate='target.foo = source.foo',
                   source_alias='source',
                   target_alias='target')
               .when_not_matched_insert_all(except_cols=["bar"])
               .execute()
            )
            {'num_source_rows': 1, 'num_target_rows_inserted': 1, 'num_target_rows_updated': 0, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 3, 'num_output_rows': 4, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas().sort_values("foo", ignore_index=True)
               foo  bar
            0  1    4
            1  2    5
            2  3    6
            3  4    NaN
            ```
        """
        maybe_source_alias = self._builder.source_alias
        maybe_target_alias = self._builder.target_alias

        src_alias = (maybe_source_alias + ".") if maybe_source_alias is not None else ""
        trgt_alias = (
            (maybe_target_alias + ".") if maybe_target_alias is not None else ""
        )

        except_columns = except_cols or []

        updates = {
            f"{trgt_alias}`{col.name}`": f"{src_alias}`{col.name}`"
            for col in self._builder.arrow_schema  # type: ignore[attr-defined]
            if col.name not in except_columns
        }

        self._builder.when_not_matched_insert(updates, predicate)
        return self

    def when_not_matched_by_source_update(
        self, updates: dict[str, str], predicate: str | None = None
    ) -> TableMerger:
        """Update a target row that has no matches in the source based on the rules defined by ``updates``.
        If a ``predicate`` is specified, then it must evaluate to true for the row to be updated.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            updates: a mapping of column name to update SQL expression.
            predicate: SQL like predicate on when to update.

        Returns:
            TableMerger: TableMerger Object

        Example:
            ```python
            from deltalake import DeltaTable, write_deltalake
            import pyarrow as pa

            data = pa.table({"x": [1, 2, 3], "y": [4, 5, 6]})
            write_deltalake("tmp", data)
            dt = DeltaTable("tmp")
            new_data = pa.table({"x": [2, 3, 4]})

            (
               dt.merge(
                   source=new_data,
                   predicate='target.x = source.x',
                   source_alias='source',
                   target_alias='target')
               .when_not_matched_by_source_update(
                   predicate = "y > 3",
                   updates = {"y": "0"})
               .execute()
            )
            {'num_source_rows': 3, 'num_target_rows_inserted': 0, 'num_target_rows_updated': 1, 'num_target_rows_deleted': 0, 'num_target_rows_copied': 2, 'num_output_rows': 3, 'num_target_files_added': 1, 'num_target_files_removed': 1, 'execution_time_ms': ..., 'scan_time_ms': ..., 'rewrite_time_ms': ...}

            dt.to_pandas().sort_values("x", ignore_index=True)
               x  y
            0  1  0
            1  2  5
            2  3  6
            ```
        """
        self._builder.when_not_matched_by_source_update(updates, predicate)
        return self

    def when_not_matched_by_source_delete(
        self, predicate: str | None = None
    ) -> TableMerger:
        """Delete a target row that has no matches in the source from the table only if the given
        ``predicate`` (if specified) is true for the target row.

        Note:
            Column names with special characters, such as numbers or spaces should be encapsulated
            in backticks: "target.`123column`" or "target.`my column`"

        Args:
            predicate:  SQL like predicate on when to delete when not matched by source.

        Returns:
            TableMerger: TableMerger Object
        """
        self._builder.when_not_matched_by_source_delete(predicate)
        return self

    def execute(self) -> dict[str, Any]:
        """Executes `MERGE` with the previously provided settings in Rust with Apache Datafusion query engine.

        Returns:
            Dict: metrics
        """
        metrics = self._table.merge_execute(self._builder)
        return json.loads(metrics)


class TableAlterer:
    """API for various table alteration commands."""

    def __init__(self, table: DeltaTable) -> None:
        self.table = table

    def add_feature(
        self,
        feature: TableFeatures | list[TableFeatures],
        allow_protocol_versions_increase: bool = False,
        commit_properties: CommitProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
    ) -> None:
        """
        Enable a table feature.

        Args:
            feature: Table Feature e.g. Deletion Vectors, Change Data Feed
            allow_protocol_versions_increase: Allow the protocol to be implicitly bumped to reader 3 or writer 7
            commit_properties: properties of the transaction commit. If None, default values are used.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable("test_table")
            dt.alter.add_feature(TableFeatures.AppendOnly)
            ```

            **Check protocol**
            ```
            dt.protocol()
            ProtocolVersions(min_reader_version=1, min_writer_version=7, writer_features=['appendOnly'], reader_features=None)
            ```
        """
        if isinstance(feature, TableFeatures):
            feature = [feature]
        self.table._table.add_feature(
            feature,
            allow_protocol_versions_increase,
            commit_properties,
            post_commithook_properties,
        )

    def add_columns(
        self,
        fields: DeltaField | list[DeltaField],
        commit_properties: CommitProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
    ) -> None:
        """Add new columns and/or update the fields of a stuctcolumn

        Args:
            fields: fields to merge into schema
            commit_properties: properties of the transaction commit. If None, default values are used.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.

        Example:
            ```python
            from deltalake.schema import Field, PrimitiveType, StructType
            dt = DeltaTable("test_table")
            new_fields = [
                Field("baz", StructType([Field("bar", PrimitiveType("integer"))])),
                Field("bar", PrimitiveType("integer"))
            ]
            dt.alter.add_columns(
                new_fields
            )
            ```
        """
        if isinstance(fields, DeltaField):
            fields = [fields]

        self.table._table.add_columns(
            fields,
            commit_properties,
            post_commithook_properties,
        )

    def add_constraint(
        self,
        constraints: dict[str, str],
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> None:
        """
        Add constraints to the table. Limited to `single constraint` at once.

        Args:
            constraints: mapping of constraint name to SQL-expression to evaluate on write
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable("test_table_constraints")
            dt.alter.add_constraint({
                "value_gt_5": "value > 5",
            })
            ```

            **Check configuration**
            ```
            dt.metadata().configuration
            {'delta.constraints.value_gt_5': 'value > 5'}
            ```
        """
        self.table._table.add_constraints(
            constraints,
            commit_properties,
            post_commithook_properties,
        )

    def drop_constraint(
        self,
        name: str,
        raise_if_not_exists: bool = True,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> None:
        """
        Drop constraints from a table. Limited to `single constraint` at once.

        Args:
            name: constraint name which to drop.
            raise_if_not_exists: set if should raise if not exists.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable("test_table_constraints")
            dt.metadata().configuration
            {'delta.constraints.value_gt_5': 'value > 5'}
            ```

            **Drop the constraint**
            ```python
            dt.alter.drop_constraint(name = "value_gt_5")
            ```

            **Configuration after dropping**
            ```python
            dt.metadata().configuration
            {}
            ```
        """
        self.table._table.drop_constraints(
            name,
            raise_if_not_exists,
            commit_properties,
            post_commithook_properties,
        )

    def set_table_properties(
        self,
        properties: dict[str, str],
        raise_if_not_exists: bool = True,
        commit_properties: CommitProperties | None = None,
    ) -> None:
        """
        Set properties from the table.

        Args:
            properties: properties which set
            raise_if_not_exists: set if should raise if not exists.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Example:
            ```python
            from deltalake import write_deltalake, DeltaTable
            import pandas as pd
            df = pd.DataFrame(
                {"id": ["1", "2", "3"],
                "deleted": [False, False, False],
                "price": [10., 15., 20.]
                })
            write_deltalake("tmp", df)

            dt = DeltaTable("tmp")
            dt.alter.set_table_properties({"delta.enableChangeDataFeed": "true"})
            ```
        """
        self.table._table.set_table_properties(
            properties,
            raise_if_not_exists,
            commit_properties,
        )

    def set_table_name(
        self,
        name: str,
        commit_properties: CommitProperties | None = None,
    ) -> None:
        """
        Set the name of the table.

        Args:
            name: the name of the table
            commit_properties: properties of the transaction commit. If None, default values are used.
                              Note: This parameter is not yet implemented and will be ignored.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable("test_table")
            dt.alter.set_table_name("new_table_name")
            ```
        """
        self.table._table.set_table_name(name, commit_properties)

    def set_table_description(
        self,
        description: str,
        commit_properties: CommitProperties | None = None,
    ) -> None:
        """
        Set the description of the table.

        Args:
            description: the description of the table
            commit_properties: properties of the transaction commit. If None, default values are used.
                              Note: This parameter is not yet implemented and will be ignored.

        Example:
            ```python
            from deltalake import DeltaTable
            dt = DeltaTable("test_table")
            dt.alter.set_table_description("new_table_description")
            ```
        """
        self.table._table.set_table_description(description, commit_properties)

    def set_column_metadata(
        self,
        column: str,
        metadata: dict[str, str],
        commit_properties: CommitProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
    ) -> None:
        """
        Update a field's metadata in a schema. If the metadata key does not exist, the entry is inserted.

        If the column name doesn't exist in the schema - an error is raised.

        :param column: name of the column to update metadata for.
        :param metadata: the metadata to be added or modified on the column.
        :param commit_properties: properties of the transaction commit. If None, default values are used.
        :param post_commithook_properties: properties for the post commit hook. If None, default values are used.
        :return:
        """
        self.table._table.set_column_metadata(
            column, metadata, commit_properties, post_commithook_properties
        )


class TableOptimizer:
    """API for various table optimization commands."""

    def __init__(self, table: DeltaTable) -> None:
        self.table = table

    def compact(
        self,
        partition_filters: FilterConjunctionType | None = None,
        target_size: int | None = None,
        max_concurrent_tasks: int | None = None,
        max_spill_size: int | None = None,
        max_temp_directory_size: int | None = None,
        min_commit_interval: int | timedelta | None = None,
        writer_properties: WriterProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """
        Compacts small files to reduce the total number of files in the table.

        This operation is idempotent; if run twice on the same table (assuming it has
        not been updated) it will do nothing the second time.

        If this operation happens concurrently with any operations other than append,
        it will fail.

        Args:
            partition_filters: the partition filters that will be used for getting the matched files
            target_size: desired file size after bin-packing files, in bytes. If not
                            provided, will attempt to read the table configuration value ``delta.targetFileSize``.
                            If that value isn't set, will use default value of 100MB.
            max_concurrent_tasks: the maximum number of concurrent tasks to use for
                                    file compaction. Defaults to number of CPUs. More concurrent tasks can make compaction
                                    faster, but will also use more memory.
            max_spill_size: the maximum number of bytes allowed in memory before spilling to disk. If not specified, uses DataFusion's default.
            max_temp_directory_size: the maximum disk space for temporary spill files. If not specified, uses DataFusion's default.
            min_commit_interval: minimum interval in seconds or as timedeltas before a new commit is
                                    created. Interval is useful for long running executions. Set to 0 or timedelta(0), if you
                                    want a commit per partition.
            writer_properties: Pass writer properties to the Rust parquet writer.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Returns:
            the metrics from optimize

        Example:
            Use a timedelta object to specify the seconds, minutes or hours of the interval.
            ```python
            from deltalake import DeltaTable, write_deltalake
            from datetime import timedelta
            import pyarrow as pa

            write_deltalake("tmp", pa.table({"x": [1], "y": [4]}))
            write_deltalake("tmp", pa.table({"x": [2], "y": [5]}), mode="append")

            dt = DeltaTable("tmp")
            time_delta = timedelta(minutes=10)
            dt.optimize.compact(min_commit_interval=time_delta)
            {'numFilesAdded': 1, 'numFilesRemoved': 2, 'filesAdded': ..., 'filesRemoved': ..., 'partitionsOptimized': 1, 'numBatches': 2, 'totalConsideredFiles': 2, 'totalFilesSkipped': 0, 'preserveInsertionOrder': True}
            ```
        """
        if isinstance(min_commit_interval, timedelta):
            min_commit_interval = int(min_commit_interval.total_seconds())

        metrics = self.table._table.compact_optimize(
            self.table._stringify_partition_values(partition_filters),
            target_size,
            max_concurrent_tasks,
            max_spill_size,
            max_temp_directory_size,
            min_commit_interval,
            writer_properties,
            commit_properties,
            post_commithook_properties,
        )
        self.table.update_incremental()
        return json.loads(metrics)

    def z_order(
        self,
        columns: Iterable[str],
        partition_filters: FilterConjunctionType | None = None,
        target_size: int | None = None,
        max_concurrent_tasks: int | None = None,
        max_spill_size: int | None = None,
        max_temp_directory_size: int | None = None,
        min_commit_interval: int | timedelta | None = None,
        writer_properties: WriterProperties | None = None,
        post_commithook_properties: PostCommitHookProperties | None = None,
        commit_properties: CommitProperties | None = None,
    ) -> dict[str, Any]:
        """
        Reorders the data using a Z-order curve to improve data skipping.

        This also performs compaction, so the same parameters as compact() apply.

        Args:
            columns: the columns to use for Z-ordering. There must be at least one column.
            partition_filters: the partition filters that will be used for getting the matched files
            target_size: desired file size after bin-packing files, in bytes. If not
                            provided, will attempt to read the table configuration value ``delta.targetFileSize``.
                            If that value isn't set, will use default value of 100MB.
            max_concurrent_tasks: the maximum number of concurrent tasks to use for
                                    file compaction. Defaults to number of CPUs. More concurrent tasks can make compaction
                                    faster, but will also use more memory.
            max_spill_size: the maximum number of bytes allowed in memory before spilling to disk. If not specified, uses DataFusion's default.
            max_temp_directory_size: the maximum disk space for temporary spill files. If not specified, uses DataFusion's default.
            min_commit_interval: minimum interval in seconds or as timedeltas before a new commit is
                                    created. Interval is useful for long running executions. Set to 0 or timedelta(0), if you
                                    want a commit per partition.
            writer_properties: Pass writer properties to the Rust parquet writer.
            post_commithook_properties: properties for the post commit hook. If None, default values are used.
            commit_properties: properties of the transaction commit. If None, default values are used.

        Returns:
            the metrics from optimize

        Example:
            Use a timedelta object to specify the seconds, minutes or hours of the interval.
            ```python
            from deltalake import DeltaTable, write_deltalake
            from datetime import timedelta
            import pyarrow as pa

            write_deltalake("tmp", pa.table({"x": [1], "y": [4]}))
            write_deltalake("tmp", pa.table({"x": [2], "y": [5]}), mode="append")

            dt = DeltaTable("tmp")
            time_delta = timedelta(minutes=10)
            dt.optimize.z_order(["x"], min_commit_interval=time_delta)
            {'numFilesAdded': 1, 'numFilesRemoved': 2, 'filesAdded': ..., 'filesRemoved': ..., 'partitionsOptimized': 0, 'numBatches': 1, 'totalConsideredFiles': 2, 'totalFilesSkipped': 0, 'preserveInsertionOrder': True}
            ```
        """
        if isinstance(min_commit_interval, timedelta):
            min_commit_interval = int(min_commit_interval.total_seconds())

        metrics = self.table._table.z_order_optimize(
            list(columns),
            self.table._stringify_partition_values(partition_filters),
            target_size,
            max_concurrent_tasks,
            max_spill_size,
            max_temp_directory_size,
            min_commit_interval,
            writer_properties,
            commit_properties,
            post_commithook_properties,
        )
        self.table.update_incremental()
        return json.loads(metrics)



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Literal

from deltalake import Schema
from deltalake._internal import Transaction as Transaction
from deltalake._internal import (
    create_table_with_add_actions as _create_table_with_add_actions,
)


@dataclass
class AddAction:
    path: str
    size: int
    partition_values: Mapping[str, str | None]
    modification_time: int
    data_change: bool
    stats: str


@dataclass(init=True)
class PostCommitHookProperties:
    """The post commit hook properties, only required for advanced usecases where you need to control this."""

    def __init__(
        self,
        create_checkpoint: bool = True,
        cleanup_expired_logs: bool | None = None,
    ) -> None:
        """Checkpoints are by default created based on the delta.checkpointInterval config setting.
        cleanup_expired_logs can be set to override the delta.enableExpiredLogCleanup, otherwise the
        config setting will be used to decide whether to clean up logs automatically by taking also
        the delta.logRetentionDuration into account.

        Args:
            create_checkpoint (bool, optional): to create checkpoints based on checkpoint interval. Defaults to True.
            cleanup_expired_logs (Optional[bool], optional): to clean up logs based on interval. Defaults to None.
        """
        self.create_checkpoint = create_checkpoint
        self.cleanup_expired_logs = cleanup_expired_logs


@dataclass(init=True)
class CommitProperties:
    """The commit properties. Controls the behaviour of the commit."""

    def __init__(
        self,
        custom_metadata: dict[str, str] | None = None,
        max_commit_retries: int | None = None,
        app_transactions: list[Transaction] | None = None,
    ) -> None:
        """Custom metadata to be stored in the commit. Controls the number of retries for the commit.

        Args:
            custom_metadata: custom metadata that will be added to the transaction commit.
            max_commit_retries: maximum number of times to retry the transaction commit.
        """
        self.custom_metadata = custom_metadata
        self.max_commit_retries = max_commit_retries
        self.app_transactions = app_transactions


def create_table_with_add_actions(
    table_uri: str,
    schema: Schema,
    add_actions: list[AddAction],
    mode: Literal["error", "append", "overwrite", "ignore"] = "error",
    partition_by: list[str] | str | None = None,
    name: str | None = None,
    description: str | None = None,
    configuration: Mapping[str, str | None] | None = None,
    storage_options: dict[str, str] | None = None,
    commit_properties: CommitProperties | None = None,
    post_commithook_properties: PostCommitHookProperties | None = None,
) -> None:
    if isinstance(partition_by, str):
        partition_by = [partition_by]
    _create_table_with_add_actions(
        table_uri=table_uri,
        schema=schema,
        add_actions=add_actions,
        mode=mode,
        partition_by=partition_by or [],
        name=name,
        description=description,
        configuration=configuration,
        storage_options=storage_options,
        commit_properties=commit_properties,
        post_commithook_properties=post_commithook_properties,
    )



import pathlib
from datetime import timedelta

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import CommitProperties, DeltaTable, write_deltalake
from deltalake.query import QueryBuilder


@pytest.mark.parametrize("use_relative", [True, False])
def test_optimize_run_table(
    tmp_path: pathlib.Path,
    sample_table: Table,
    monkeypatch,
    use_relative: bool,
):
    if use_relative:
        monkeypatch.chdir(tmp_path)  # Make tmp_path the working directory
        (tmp_path / "path/to/table").mkdir(parents=True)
        table_path = "./path/to/table"
    else:
        table_path = str(tmp_path)

    write_deltalake(table_path, sample_table, mode="append")
    write_deltalake(table_path, sample_table, mode="append")
    write_deltalake(table_path, sample_table, mode="append")

    dt = DeltaTable(table_path)
    old_data = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id")
        .read_all()
    )
    old_version = dt.version()

    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.optimize.compact(commit_properties=commit_properties)

    new_data = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id")
        .read_all()
    )
    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    assert last_action["userName"] == "John Doe"
    assert dt.version() == old_version + 1
    assert old_data == new_data


def test_z_order_optimize(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )

    dt = DeltaTable(tmp_path)
    old_version = dt.version()

    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.optimize.z_order(["sold", "price"], commit_properties=commit_properties)
    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    assert last_action["userName"] == "John Doe"
    assert dt.version() == old_version + 1
    assert len(dt.file_uris()) == 1


def test_optimize_min_commit_interval(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    write_deltalake(tmp_path, sample_table, partition_by="id", mode="append")
    write_deltalake(tmp_path, sample_table, partition_by="id", mode="append")
    write_deltalake(tmp_path, sample_table, partition_by="id", mode="append")

    dt = DeltaTable(tmp_path)
    old_version = dt.version()

    dt.optimize.z_order(["sold", "price"], min_commit_interval=timedelta(0))

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    # The table has 5 distinct partitions, each of which are Z-ordered
    # independently. So with min_commit_interval=0, each will get its
    # own commit.
    assert dt.version() == old_version + 5


def test_optimize_schema_evolved_table(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    data = Table(
        {
            "foo": Array(
                ["1"],
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append", schema_mode="merge")

    data = Table(
        {
            "bar": Array(
                ["1"],
                ArrowField("bar", type=DataType.string(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append", schema_mode="merge")

    dt = DeltaTable(tmp_path)
    old_version = dt.version()

    dt.optimize.compact()

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    assert dt.version() == old_version + 1

    data = Table(
        {
            "foo": Array(
                ["1", None],
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
            "bar": Array(
                [None, "1"],
                ArrowField("bar", type=DataType.string(), nullable=True),
            ),
        }
    )

    assert (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by foo asc")
        .read_all()
        == data
    )


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_zorder_with_space_partition(tmp_path: pathlib.Path):
    import pandas as pd

    df = pd.DataFrame(
        {
            "user": ["James", "Anna", "Sara", "Martin"],
            "country": ["United States", "Canada", "Costa Rica", "South Africa"],
            "age": [34, 23, 45, 26],
        }
    )

    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        mode="overwrite",
        partition_by=["country"],
    )

    test_table = DeltaTable(tmp_path)

    # retrieve by partition works fine
    partitioned_df = test_table.to_pandas(
        partitions=[("country", "=", "United States")],
    )
    print(partitioned_df)

    test_table.optimize.z_order(columns=["user"])


@pytest.mark.pyarrow
def test_optimize_schema_evolved_3185(tmp_path):
    """https://github.com/delta-io/delta-rs/issues/3185"""
    import pyarrow as pa

    # Define the data for the first write
    data_first_write = pa.array(
        [
            {"name": "Alice", "age": 30, "details": {"email": "alice@example.com"}},
            {"name": "Bob", "age": 25, "details": {"email": "bob@example.com"}},
        ]
    )

    data_second_write = pa.array(
        [
            {
                "name": "Charlie",
                "age": 35,
                "details": {"address": "123 Main St", "email": "charlie@example.com"},
            },
            {
                "name": "Diana",
                "age": 28,
                "details": {"address": "456 Elm St", "email": "diana@example.com"},
            },
        ]
    )

    schema_first_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            ("details", pa.struct([("email", pa.string())])),
        ]
    )

    schema_second_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            (
                "details",
                pa.struct(
                    [
                        ("address", pa.string()),
                        ("email", pa.string()),
                    ]
                ),
            ),
        ]
    )
    table_first_write = pa.Table.from_pylist(
        data_first_write, schema=schema_first_write
    )
    table_second_write = pa.Table.from_pylist(
        data_second_write, schema=schema_second_write
    )

    write_deltalake(
        tmp_path,
        table_first_write,
        mode="append",
    )

    write_deltalake(tmp_path, table_second_write, mode="append", schema_mode="merge")

    dt = DeltaTable(tmp_path)

    dt.optimize.z_order(columns=["name"])

    assert dt.version() == 2
    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"


def test_compact_with_spill_parameters(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    write_deltalake(tmp_path, sample_table, mode="append")
    write_deltalake(tmp_path, sample_table, mode="append")
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)
    old_version = dt.version()
    old_num_files = len(dt.file_uris())

    dt.optimize.compact(
        max_spill_size=100 * 1024 * 1024 * 1024,  # 100 GB
        max_temp_directory_size=500 * 1024 * 1024 * 1024,  # 500 GB
    )

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    assert dt.version() == old_version + 1
    assert len(dt.file_uris()) <= old_num_files


def test_z_order_with_spill_parameters(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    write_deltalake(tmp_path, sample_table, mode="append")
    write_deltalake(tmp_path, sample_table, mode="append")
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)
    old_version = dt.version()

    dt.optimize.z_order(
        columns=["sold", "price"],
        max_spill_size=100 * 1024 * 1024 * 1024,  # 100 GB
        max_temp_directory_size=500 * 1024 * 1024 * 1024,  # 500 GB
    )

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "OPTIMIZE"
    assert dt.version() == old_version + 1
    assert len(dt.file_uris()) == 1
